# Copyright 2019,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
HMC definitions for zhmcclient end2end tests.
"""


import os

from ._hmc_inventory_file import HMCInventoryFile
from ._hmc_vault_file import HMCVaultFile
from ._hmc_definition import HMCDefinition

__all__ = ['print_hmc_definitions', 'hmc_definitions', 'HMCDefinitions',
           'HMCNoVaultError', 'HMCNotFound']

# Path name of user's home directory
HOME_DIR = os.path.expanduser("~")

# Default file names of HMC inventory and vault files in user's home directory.
# Keep in sync with Makefile.
DEFAULT_TESTINVENTORY_FN = '.zhmc_inventory.yaml'
DEFAULT_TESTVAULT_FN = '.zhmc_vault.yaml'

DEFAULT_TESTINVENTORY_FILE = os.path.join(HOME_DIR, DEFAULT_TESTINVENTORY_FN)
DEFAULT_TESTVAULT_FILE = os.path.join(HOME_DIR, DEFAULT_TESTVAULT_FN)

# Default group name or HMC nickname in HMC inventory file to test against.
# Keep in sync with Makefile.
DEFAULT_TESTHMC = 'default'


def hmc_definitions(load=True):
    """
    Return the list of HMC definitions to be used for testing.

    Parameters:

      load (bool): Load the HMC inventory and vault files. Otherwise, these
        files are not loaded. This is used to avoid a dependency on these
        files for normal zhmcclient users.
        If `None`, the 'TESTEND2END_LOAD' environment variable is used.

    Returns:

      list of :class:`zhmcclient.testutils.HMCDefinition`
    """
    hmcdefs = HMCDefinitions(load=load)
    hd_list = hmcdefs.list_hmcs(hmcdefs.testhmc)
    return hd_list


def print_hmc_definitions():
    """
    Print all available HMC definitions in the HMC inventory and vault files.
    """
    hmcdefs = HMCDefinitions()
    print("\nHMC definitions for end2end tests:")

    print(f"\nHMC inventory file: {hmcdefs.inventory_file}")
    print(f"HMC vault file: {hmcdefs.vault_file}")
    print(f"Default test group/nickname: {hmcdefs.testhmc}")

    print("\nHMCs in inventory file:")
    print("{:20s} {:24s} {:}".
          format("HMC nickname", "HMC host / mock file", "Description"))
    for hd in hmcdefs.list_all_hmcs():
        host = hd.mock_file or hd.host
        if isinstance(host, list):
            host = ','.join(host)
        print(f"{hd.nickname:20s} {str(host):24s} {hd.description}")

    print("\nGroups in inventory file:")
    print("Group name           HMCs in the group")
    for group_name in hmcdefs.list_all_group_names():
        hmc_names = ', '.join(
            [hd.nickname for hd in hmcdefs.list_hmcs(group_name)])
        print(f"{group_name:20s} {hmc_names}")


def _default(vars_tuple, key, default):
    """
    Return the key value from one of the dicts in vars_tuple, or the default.
    """
    for vars_dict in vars_tuple:
        if vars_dict and key in vars_dict:
            return vars_dict[key]
    return default


class HMCNoVaultError(Exception):
    """
    The :ref:`HMC vault file` does not have a corresponding entry for the HMC.
    """
    pass


class HMCNotFound(Exception):
    """
    The HMC group or nickname was not found in the :ref:`HMC inventory file`.
    """
    pass


class HMCDefinitions:
    """
    The HMC definitions in the :ref:`HMC inventory file` and their credentials
    in the :ref:`HMC vault file`.
    """

    def __init__(self, inventory_file=None, vault_file=None, testhmc=None,
                 load=None):
        """
        Parameters:

          inventory_file (string): Path name of HMC inventory file`.
            If `None`, the file specified in the 'TESTINVENTORY' environment
            variable or if not set, the default file ``~/.zhmc_inventory.yaml``
            is used.

          vault_file (string): Path name of HMC vault file.
            If `None`, the file specified in the 'TESTVAULT' environment
            variable or if not set, the default file ``~/.zhmc_vault.yaml``
            is used.

          testhmc (string): Group nickname or HMC nickname in HMC inventory file
            to test against.
            If `None`, the nickname specified in the 'TESTHMC' environment
            variable or if not set, the nickname "default" is used.

          load (bool): Load the HMC inventory and vault files. Otherwise, these
            files are not loaded. This is used to avoid a dependency on these
            files for normal zhmcclient users.
            If `None`, the 'TESTEND2END_LOAD' environment variable is used.

        Raises:

          zhmcclient.testutils.HMCInventoryFileError:
          zhmcclient.testutils.HMCVaultFileError:
          zhmcclient.testutils.HMCNoVaultError:
        """

        # The Sphinx build imports this module and the use of this function
        # in the hmc_definition() fixture along with the wildcard imports in
        # the testutils/__init__.py module causes this function to be executed
        # upon module import. Since there are no HMC inventory and vault files
        # when GitHub Actions or ReadTheDocs builds the documentation, the
        # boolean 'TESTEND2END_LOAD' env.var is used to enable the loading of
        # the files in these cases. This env.var needs to be set to 'true' in
        # the Makefile for end2end targets.
        if load is None:
            load = bool(os.getenv('TESTEND2END_LOAD'))

        if not load:
            self._inventory_file = None
            self._vault_file = None
            self._testhmc = None
            self._inventory = None
            self._vault = None
            self._hd_dict = {}
        else:
            self._inventory_file = inventory_file or os.getenv(
                'TESTINVENTORY', DEFAULT_TESTINVENTORY_FILE)
            self._vault_file = vault_file or os.getenv(
                'TESTVAULT', DEFAULT_TESTVAULT_FILE)
            self._testhmc = testhmc or os.getenv(
                'TESTHMC', DEFAULT_TESTHMC)

            self._inventory = HMCInventoryFile(self._inventory_file)
            self._vault = HMCVaultFile(self._vault_file)

            # All HMC definitions, represented as a flat set of groups
            # with their hosts. All variables are stored in the hosts.
            # The list methods create and return HMCDefinition objects
            # from this representation.
            # Dictionary structure:
            #   key: group name, value: dict of HMC definitions with:
            #     key: HMC nickname, value: dict of HMCDefinition attrs with:
            #       key: attr name, value: attr value
            self._hd_dict = {}

            # Update self._hd_dict from groups, starting with top-level groups
            self._init_groups(self._inventory.data, {})

            # Update self._hd_dict with propagated host properties
            self._propagate_groups(self._inventory.data)

            # Update self._hd_dict from HMC vault file
            self._add_vault_data(self._vault, self._inventory.filepath)

    def _init_groups(self, groups, inherited_vars):
        """
        Process groups from an HMC inventory filer.

        Parameters:

          groups (dict): The groups, as a data structure from the HMC inventory
            file. Key: group name, Value: dict with hosts/vars/children.
            At the top level, groups is a dict with a single item 'all'. Below
            that, groups is a dict with the items from the 'children' item.

          inherited_vars (dict): Additional variables inherited from parents.
        """
        for group_name, group_dict in groups.items():
            if not group_dict:
                continue
            group_vars = dict(inherited_vars)
            group_vars.update(group_dict.get('vars') or {})
            self._init_group(group_name, group_dict, group_vars)
            child_groups = group_dict.get('children') or {}
            self._init_groups(child_groups, group_vars)

    def _init_group(self, group_name, group_dict, inherited_vars):
        """
        Process a single group from an HMC inventory file.

        Parameters:

          group_name (string): Name of the group.

          group_dict (dict): The group, as a dict with hosts/vars/children.

          inherited_vars (dict): Additional variables inherited from parents.
        """

        _host_dict = {}
        hosts = group_dict.get('hosts') or {}
        for nickname, host_vars in hosts.items():

            combined_vars = dict(inherited_vars)
            if host_vars:
                combined_vars.update(host_vars)

            description = combined_vars.pop('description', '')
            contact = combined_vars.pop('contact', '')
            access_via = combined_vars.pop('access_via', '')
            mock_file = combined_vars.pop('mock_file', None)
            ansible_host = combined_vars.pop('ansible_host', None)
            cpcs = combined_vars.pop('cpcs', None)
            add_vars = combined_vars

            _host_dict[nickname] = dict(
                description=description, contact=contact, access_via=access_via,
                mock_file=mock_file, ansible_host=ansible_host, cpcs=cpcs,
                add_vars=add_vars)

        self._hd_dict[group_name] = _host_dict

    def _propagate_groups(self, groups, parent_hosts=None):
        """
        Propagate host properties of groups to their child groups.

        Parameters:

          groups (dict): The groups, as a data structure from the HMC inventory
            file. Key: group name, Value: dict with hosts/vars/children.
            At the top level, groups is a dict with a single item 'all'. Below
            that, groups is a dict with the items from the 'children' item.

          parent_hosts (dict): The parent hosts of the groups, as a data
            structure from the HMC inventory file, or None for top level.
            Key: host nickname, Value: dict with host attributes.
        """
        for group_name, group_dict in groups.items():

            if not group_dict:
                continue

            self._propagate_group(group_name, group_dict, parent_hosts)

            child_groups = group_dict.get('children') or {}
            hosts = group_dict.get('hosts') or {}
            self._propagate_groups(child_groups, hosts)

    def _propagate_group(self, group_name, group_dict, parent_hosts):
        """
        Propagate host properties of a group to its hosts.

        Parameters:

          group_name (string): Name of the group.

          group_dict (dict): The group, as a dict with hosts/vars/children.

          parent_hosts (dict): The parent hosts of the group, as a data
            structure from the HMC inventory file, or None for top level.
            Key: host nickname, Value: dict with host attributes.
        """
        if parent_hosts is None:
            parent_hosts = {}

        hosts = group_dict.get('hosts') or {}
        for nickname, host_vars in hosts.items():

            try:
                parent_vars = parent_hosts[nickname]
            except KeyError:
                parent_vars = {}

            if host_vars is not None:
                host_vars_new = dict(host_vars)
            else:
                host_vars_new = {}

            if 'description' in parent_vars:
                host_vars_new.setdefault(
                    'description', parent_vars['description'])
            if 'contact' in parent_vars:
                host_vars_new.setdefault(
                    'contact', parent_vars['contact'])
            if 'access_via' in parent_vars:
                host_vars_new.setdefault(
                    'access_via', parent_vars['access_via'])
            if 'ansible_host' in parent_vars:
                host_vars_new.setdefault(
                    'ansible_host', parent_vars['ansible_host'])
            if 'mock_file' in parent_vars:
                host_vars_new.setdefault(
                    'mock_file', parent_vars['mock_file'])
            if 'cpcs' in parent_vars:
                host_vars_new.setdefault(
                    'cpcs', parent_vars['cpcs'])
            if 'add_vars' in parent_vars:
                host_vars_new.setdefault(
                    'add_vars', parent_vars['add_vars'])

            # The vault file variables are not in parent_vars and cannot
            # be propagated (they are defined once in the vault file).

            self._hd_dict[group_name][nickname].update(host_vars_new)

    def _add_vault_data(self, hmc_vault, inv_file):
        """
        Add data from HMC vault file to each host of all groups.

        Parameters:

          hmc_vault (HMCVault): Content of HMC vault file.

          inv_file (string): Path name of HMV inventory file (for messages).

        Raises:

          HMCNoVaultError: No cprresponding entry in HMC vault file.
        """
        auth = hmc_vault.data['hmc_auth'] or {}
        for group_dict in self._hd_dict.values():
            for nickname, host_vars in group_dict.items():

                try:
                    auth_vars = auth[nickname]
                except KeyError:
                    new_exc = HMCNoVaultError(
                        f"HMC {nickname!r} defined in HMC inventory file "
                        f"{inv_file} has no corresponding entry in HMC "
                        f"vault file {hmc_vault.filepath}")
                    new_exc.__cause__ = None
                    raise new_exc  # HMCNoVaultError

                # userid and password are required by the JSON schema
                host_vars['userid'] = auth_vars['userid']
                host_vars['password'] = auth_vars['password']
                host_vars['verify'] = auth_vars.get('verify', True)
                host_vars['ca_certs'] = auth_vars.get('ca_certs', None)

    def __repr__(self):
        return (
            "HMCDefinitions("
            f"inventory_file={self.inventory_file!r}, "
            f"vault_file={self.vault_file!r}, "
            f"testhmc={self.testhmc!r}, "
            f"group_names={self.group_names!r})")

    @property
    def inventory_file(self):
        """
        string: Path name of HMC inventory file.
        """
        return self._inventory_file

    @property
    def vault_file(self):
        """
        string: Path name of HMC vault file.
        """
        return self._vault_file

    @property
    def testhmc(self):
        """
        string: HMC group or single HMC nickname to be tested.
        """
        return self._testhmc

    @property
    def group_names(self):
        """
        list of string: The names of the HMC groups in the HMC inventory file.
        """
        return list(self._hd_dict.keys())

    def list_hmcs(self, name=None):
        """
        Return a list of :class:`HMCDefinition` objects in the HMC inventory
        file, for the specified HMC group or single HMC nickname.

        The path names in the :attr:`HMCDefinition.mock_file` property are
        absolute path names, whereby any relative path names in the HMC
        inventory file have been interpreted relative to the directory of
        the HMC inventory file.

        Parameters:

          name (string): Name of an HMC group or nickname of a single HMC in the
            HMC inventory file. If `None`, the default group or nickname
            defined in :attr:`HMCDefinitions.testhmc` is used.

        Returns:

          list of :class:`HMCDefinition`: The specified HMCs in the HMC
            inventory file.

        Raises:

          zhmcclient.testutils.HMCNotFound: HMC group or nickname not found.
        """
        if not self._hd_dict:
            return []

        if name is None:
            name = self._testhmc

        # Try to find a group with this name
        try:
            host_dict = self._hd_dict[name]
        except KeyError:
            # Try to find a single HMC with this nickname in group 'all'
            all_host_dict = self._hd_dict['all']
            for nickname, host_vars in all_host_dict.items():
                if nickname == name:
                    host_dict = {name: host_vars}
                    break
            else:
                new_exc = HMCNotFound(
                    f"HMC group or nickname {name!r} not found in HMC "
                    f"inventory file {self._inventory_file}")
                new_exc.__cause__ = None
                raise new_exc  # HMCNotFound

        hd_list = []
        for nickname, host_vars in host_dict.items():

            description = host_vars.get('description', '')
            contact = host_vars.get('contact', '')
            access_via = host_vars.get('access_via', '')
            mock_file = host_vars.get('mock_file', None)
            userid = host_vars.get('userid', None)
            password = host_vars.get('password', None)
            verify = host_vars.get('verify', None)
            ca_certs = host_vars.get('ca_certs', None)
            cpcs = host_vars.get('cpcs', None)
            add_vars = host_vars.get('add_vars', None)

            if mock_file:
                # The host will be set from the 'host' attribute in the
                # mock file. Since that file is read only later, we set
                # the HMC definition's host to None, for now.
                host = None
            else:
                # If ansible_host was set, it is always used, and otherwise
                # the HMC nickname is used as a DNS name or IP address.
                ansible_host = host_vars.get('ansible_host')
                # Note: ansible_host may be a string or list of strings.
                host = ansible_host or nickname

            # Make relative mock_file relative to inventory file
            if mock_file and not os.path.isabs(mock_file):
                mock_file = os.path.join(
                    os.path.dirname(self._inventory_file), mock_file)

            hd = HMCDefinition(
                nickname=nickname, description=description,
                contact=contact, access_via=access_via, mock_file=mock_file,
                host=host, userid=userid, password=password, verify=verify,
                ca_certs=ca_certs, cpcs=cpcs, add_vars=add_vars)

            hd_list.append(hd)

        return hd_list

    def list_all_hmcs(self):
        """
        List all HMCs in the HMC inventory file.

        Returns:

          list of :class:`HMCDefinition`: All HMCs in the HMC inventory file.
        """
        return self.list_hmcs('all')

    def list_all_group_names(self):
        """
        List all group names in the HMC inventory file.

        Returns:

          list of string: All group names in the HMC inventory file.
        """
        return list(self._hd_dict.keys())



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
End2end tests for HBAs (on CPCs in DPM mode without storage mgmt feature).

These tests do not change any existing partitions or HBAs, but create, modify
and delete test partitions with HBAs.
"""


import warnings
import pytest
from requests.packages import urllib3

import zhmcclient
# pylint: disable=line-too-long,unused-import
from zhmcclient.testutils import hmc_definition, hmc_session  # noqa: F401, E501
from zhmcclient.testutils import dpm_mode_cpcs  # noqa: F401, E501
# pylint: enable=line-too-long,unused-import

from .utils import skip_warn, pick_test_resources, TEST_PREFIX, \
    standard_partition_props, skipif_storage_mgmt_feature, \
    runtest_find_list, runtest_get_properties

urllib3.disable_warnings()

# Properties in minimalistic Hba objects (e.g. find_by_name())
HBA_MINIMAL_PROPS = ['element-uri', 'name']

# Properties in Hba objects returned by list() without full props
HBA_LIST_PROPS = ['element-uri', 'name', 'description', 'wwpn']

# Properties whose values can change between retrievals of Hba objects
HBA_VOLATILE_PROPS = []


def test_hba_find_list(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test list(), find(), findall().
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_storage_mgmt_feature(cpc)

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the HBAs to test with
        part_hba_tuples = []
        part_list = cpc.partitions.list()
        for part in part_list:
            hba_list = part.hbas.list()
            for hba in hba_list:
                part_hba_tuples.append((part, hba))
        if not part_hba_tuples:
            skip_warn(
                f"No partitions with HBAs on CPC {cpc.name} managed by HMC "
                f"{hd.host}")
        part_hba_tuples = pick_test_resources(part_hba_tuples)

        for part, hba in part_hba_tuples:
            print(f"Testing on CPC {cpc.name} with HBA {hba.name!r} of "
                  f"partition {part.name!r}")
            runtest_find_list(
                session, part.hbas, hba.name, 'name', 'wwpn',
                HBA_VOLATILE_PROPS, HBA_MINIMAL_PROPS, HBA_LIST_PROPS)


def test_hba_property(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test property related methods
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled

        session = cpc.manager.session
        hd = session.hmc_definition

        # Pick the HBAs to test with
        part_hba_tuples = []
        part_list = cpc.partitions.list()
        for part in part_list:
            hba_list = part.hbas.list()
            for hba in hba_list:
                part_hba_tuples.append((part, hba))
        if not part_hba_tuples:
            skip_warn(
                f"No partitions with HBAs on CPC {cpc.name} managed by "
                f"HMC {hd.host}")
        part_hba_tuples = pick_test_resources(part_hba_tuples)

        for part, hba in part_hba_tuples:
            print(f"Testing on CPC {cpc.name} with HBA {hba.name!r} of "
                  f"partition {part.name!r}")

            # Select a property that is not returned by list()
            non_list_prop = 'description'

            runtest_get_properties(hba.manager, non_list_prop)


def test_hba_crud(dpm_mode_cpcs):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Test create, read, update and delete a HBA (and a partition).
    """
    if not dpm_mode_cpcs:
        pytest.skip("HMC definition does not include any CPCs in DPM mode")

    for cpc in dpm_mode_cpcs:
        assert cpc.dpm_enabled
        skipif_storage_mgmt_feature(cpc)

        print(f"Testing on CPC {cpc.name}")

        part_name = TEST_PREFIX + ' test_hba_crud part1'
        hba_name = 'hba1'
        hba_name_new = hba_name + ' new'

        # Ensure a clean starting point for this test
        try:
            part = cpc.partitions.find(name=part_name)
        except zhmcclient.NotFound:
            pass
        else:
            warnings.warn(
                f"Deleting test partition from previous run: {part_name!r} on "
                f"CPC {cpc.name}", UserWarning)
            status = part.get_property('status')
            if status != 'stopped':
                part.stop()
            part.delete()

        part = None
        try:

            # Pick a FICON adapter backing the HBA
            adapters = cpc.adapters.findall(**{'type': 'fcp'})
            assert len(adapters) >= 1, (
                f"CPC {cpc.name} does not have any FCP-type FICON adapters")
            adapter = adapters[-1]  # Pick the last adapter found
            port = adapter.ports.list()[0]  # Pick its first port

            # Create a partition that will lateron contain the HBA
            part_props = standard_partition_props(cpc, part_name)
            part = cpc.partitions.create(part_props)

            # Test creating a HBA

            hba_input_props = {
                'name': hba_name,
                'description': 'Dummy HBA description.',
                'adapter-port-uri': port.uri,
                'device-number': '0100',
            }
            hba_auto_props = {}

            # The code to be tested
            hba = part.hbas.create(hba_input_props)

            for pn, exp_value in hba_input_props.items():
                assert hba.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            hba.pull_full_properties()
            for pn, exp_value in hba_input_props.items():
                assert hba.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"
            for pn, exp_value in hba_auto_props.items():
                assert hba.properties[pn] == exp_value, \
                    f"Unexpected value for property {pn!r}"

            # Test updating a property of the HBA

            new_desc = "Updated HBA description."

            # The code to be tested
            hba.update_properties(dict(description=new_desc))

            assert hba.properties['description'] == new_desc
            hba.pull_full_properties()
            assert hba.properties['description'] == new_desc

            # Test renaming the HBA

            # The code to be tested
            hba.update_properties(dict(name=hba_name_new))

            assert hba.properties['name'] == hba_name_new
            hba.pull_full_properties()
            assert hba.properties['name'] == hba_name_new
            with pytest.raises(zhmcclient.NotFound):
                part.hbas.find(name=hba_name)

            # Test deleting the HBA

            # The code to be tested
            hba.delete()

            with pytest.raises(zhmcclient.NotFound):
                part.hbas.find(name=hba_name_new)

        finally:
            # Cleanup
            if part:
                part.delete()



# Copyright 2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Group` is a user-defined group of resources that are managed by the
HMC.

Groups can be used for example in User Role permissions.

Groups can be one of two kinds:

* pattern-matching groups: Membership is implicit and the result of pattern
  matching on the resource name. This is supported only for members that are
  CPCs, Logical Partitions (i.e. classic mode), or other Groups.

* explicit groups: Each member is added and removed explicitly, using the
  :meth:`Group.add_member` and :meth:`Group.remove_member` methods.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_GROUP

__all__ = ['GroupManager', 'Group']


class GroupManager(BaseManager):
    """
    Manager providing access to the :term:`Groups <Group>` exposed by the
    HMC this client is connected to.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    HMC/SE version requirements: None
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #      Console object representing the HMC.

        # Resource properties that are supported as filter query parameters
        # (for server-side filtering).
        query_props = [
            'name',
        ]

        super().__init__(
            resource_class=Group,
            class_name=RC_GROUP,
            session=console.manager.session,
            parent=console,
            base_uri='/api/groups',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: :term:`Console` defining the scope for
        this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the Groups managed by the HMC this client is connected to.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to any Group to be included in the result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Group` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'groups'
        list_uri = '/api/groups'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args, None)

    @logged_api_call
    def create(self, properties):
        """
        Create a Group in the HMC.

        HMC/SE version requirements: None

        Authorization requirements:

        * Task permission to the "Grouping" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Custom Group' in the :term:`HMC API` book.

        Returns:

          Group:
            The resource object for the new Group.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self._base_uri, body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        group = Group(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return group


class Group(BaseResource):
    """
    Representation of a :term:`Group`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.GroupManager`).

    Note that Group objects do not have any writeable properties, so they
    do not have an ``update_properties()`` method.

    HMC/SE version requirements: None
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.GroupManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, GroupManager), (
            f"Group init: Expected manager type {GroupManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    def dump(self):
        """
        Dump this Group resource with its properties as a resource definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        return resource_dict

    # Note: Group resources cannot have their properties updated,
    #       hence there is no update_properties() method.

    @logged_api_call
    def delete(self):
        """
        Delete this Group.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this Group.
        * Task permission to the "Grouping" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def add_member(self, uri):
        """
        Add a resource as a new member to the Group.

        The group must be defined without a pattern-matching specification.
        The member resource must not already be a member of this group.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this Group.
        * Object-access permission to the resource that is added as a member.
        * Task permission to the "Grouping" task.

        Parameters:

          uri (:term:`string`): URI of the resource to be added.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'object-uri': uri}
        ops_uri = self.uri + '/operations/add-member'
        self.manager.session.post(ops_uri, resource=self, body=body)

    @logged_api_call
    def remove_member(self, uri):
        """
        Remove a member resource from the Group.

        The group must be defined without a pattern-matching specification.
        The resource must be a member of this group.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this Group.
        * Object-access permission to the resource that is added as a member.
        * Task permission to the "Grouping" task.

        Parameters:

          uri (:term:`string`): URI of the resource to be removed.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'object-uri': uri}
        ops_uri = self.uri + '/operations/remove-member'
        self.manager.session.post(ops_uri, resource=self, body=body)

    @logged_api_call
    def list_members(self):
        """
        List Group members.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this Group.
        * Object-access permission to each member resource of the Group.

        Returns:

          list: List of members.
          Each list item is a dictionary with items:

          - "object-uri" (term:`string`): URI of the member resource.
          - "name" (term:`string`): Name of the member resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`

        """
        uri = self.uri + '/members'
        result = self.manager.session.get(uri, resource=self)
        return result['members']



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _adapter module.
"""


import copy
import re
import pytest

from zhmcclient import Client, Adapter, NotFound, HTTPError, CeasedExistence
from zhmcclient_mock import FakedSession
from tests.common.utils import assert_resources


# Object IDs and names of our faked adapters:
OSA1_OID = 'osa1-oid'
OSA1_NAME = 'osa 1'
HS2_OID = 'hs2-oid'
HS2_NAME = 'hs 2'


class TestAdapter:
    """All tests for the Adapter and AdapterManager classes."""

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.

        Set up a faked session, and add a faked CPC in DPM mode without any
        child resources.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = FakedSession('fake-host', 'fake-hmc', '2.13.1', '1.8')
        self.client = Client(self.session)

        self.faked_cpc = self.session.hmc.cpcs.add({
            'object-id': 'fake-cpc1-oid',
            # object-uri is set up automatically
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc1-name',
            'description': 'CPC #1 (DPM mode)',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
            'machine-type': '2964',  # z13
            'machine-model': 'N10',
        })
        self.cpc = self.client.cpcs.find(name='fake-cpc1-name')

    def add_standard_osa(self):
        """Add a standard OSA adapter with one port to the faked HMC."""

        # Adapter properties that will be auto-set:
        # - object-uri
        # - adapter-family
        # - network-port-uris (to empty array)
        faked_osa1 = self.faked_cpc.adapters.add({
            'object-id': OSA1_OID,
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': OSA1_NAME,
            'description': 'OSA #1',
            'status': 'active',
            'type': 'osd',
            'adapter-id': '123',
            'detected-card-type': 'osa-express-5s-10gb',
            'card-location': '1234-5678-J.01',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 80,
            'used-capacity': 0,
            'allowed-capacity': 80,
            'maximum-total-capacity': 80,
            'physical-channel-status': 'operating',
        })

        # Port properties that will be auto-set:
        # - element-uri
        # Properties in parent adapter that will be auto-set:
        # - network-port-uris
        faked_osa1.ports.add({
            'element-id': 'fake-port11-oid',
            'parent': faked_osa1.uri,
            'class': 'network-port',
            'index': 0,
            'name': 'fake-port11-name',
            'description': 'OSA #1 Port #1',
        })
        return faked_osa1

    def add_standard_hipersocket(self):
        """Add a standard Hipersocket adapter with one port to the faked
        HMC."""

        # Adapter properties that will be auto-set:
        # - object-uri
        # - adapter-family
        # - network-port-uris (to empty array)
        faked_hs2 = self.faked_cpc.adapters.add({
            'object-id': HS2_OID,
            'parent': self.faked_cpc.uri,
            'class': 'adapter',
            'name': HS2_NAME,
            'description': 'Hipersocket #2',
            'status': 'active',
            'type': 'hipersockets',
            'adapter-id': '123',
            'detected-card-type': 'hipersockets',
            'port-count': 1,
            'network-port-uris': [],
            'state': 'online',
            'configured-capacity': 32,
            'used-capacity': 0,
            'allowed-capacity': 32,
            'maximum-total-capacity': 32,
            'physical-channel-status': 'operating',
            'maximum-transmission-unit-size': 56,
        })

        # Port properties that will be auto-set:
        # - element-uri
        # Properties in parent adapter that will be auto-set:
        # - network-port-uris
        faked_hs2.ports.add({
            'element-id': 'fake-port21-oid',
            'parent': faked_hs2.uri,
            'class': 'network-port',
            'index': 0,
            'name': 'fake-port21-name',
            'description': 'Hipersocket #2 Port #1',
        })
        return faked_hs2

    @staticmethod
    def add_crypto_ce5s(faked_cpc):
        """Add a Crypto Express 5S adapter to a faked CPC."""

        # Adapter properties that will be auto-set:
        # - object-uri
        # - adapter-family
        faked_adapter = faked_cpc.adapters.add({
            'object-id': 'fake-ce5s-oid',
            'parent': faked_cpc.uri,
            'class': 'adapter',
            'name': 'fake-ce5s-name',
            'description': 'Crypto Express 5S #1',
            'status': 'active',
            'type': 'crypto',
            'adapter-id': '123',
            'detected-card-type': 'crypto-express-5s',
            'card-location': 'vvvv-wwww',
            'state': 'online',
            'physical-channel-status': 'operating',
            'crypto-number': 7,
            'crypto-type': 'ep11-coprocessor',
            'udx-loaded': False,
            'tke-commands-enabled': False,
        })
        return faked_adapter

    @staticmethod
    def add_ficon_fe6sp(faked_cpc):
        """Add a not-configured FICON Express 6S+ adapter to a faked CPC."""

        # Adapter properties that will be auto-set:
        # - object-uri
        # - storage-port-uris
        faked_ficon_adapter = faked_cpc.adapters.add({
            'object-id': 'fake-ficon6s-oid',
            'parent': faked_cpc.uri,
            'class': 'adapter',
            'name': 'fake-ficon6s-name',
            'description': 'FICON Express 6S+ #1',
            'status': 'active',
            'type': 'not-configured',
            'adapter-id': '124',
            'adapter-family': 'ficon',
            'detected-card-type': 'ficon-express-16s-plus',
            'card-location': 'vvvv-wwww',
            'port-count': 1,
            'state': 'online',
            'configured-capacity': 254,
            'used-capacity': 0,
            'allowed-capacity': 254,
            'maximum-total-capacity': 254,
            'channel-path-id': None,
            'physical-channel-status': 'not-defined',
        })

        # Port properties that will be auto-set:
        # - element-uri
        # Properties in parent adapter that will be auto-set:
        # - storage-port-uris
        faked_ficon_adapter.ports.add({
            'element-id': 'fake-port11-oid',
            'parent': faked_ficon_adapter.uri,
            'class': 'storage-port',
            'index': 0,
            'name': 'fake-port11-name',
            'description': 'FICON #1 Port #1',
        })
        return faked_ficon_adapter

    def add_cpc_z13s(self):
        """Add a CPC #2 of type z13s to the faked HMC."""

        # CPC properties that will be auto-set:
        # - object-uri
        faked_cpc = self.session.hmc.cpcs.add({
            'object-id': 'fake-cpc-2-oid',
            'parent': None,
            'class': 'cpc',
            'name': 'fake-cpc-2-name',
            'description': 'CPC z13s #2',
            'status': 'active',
            'dpm-enabled': True,
            'is-ensemble-member': False,
            'iml-mode': 'dpm',
            'machine-type': '2965',  # z13s
            'machine-model': 'N10',
        })
        return faked_cpc

    def test_adaptermanager_initial_attrs(self):
        """Test initial attributes of AdapterManager."""

        adapter_mgr = self.cpc.adapters

        # Verify all public properties of the manager object
        assert adapter_mgr.resource_class == Adapter
        assert adapter_mgr.session == self.session
        assert adapter_mgr.parent == self.cpc
        assert adapter_mgr.cpc == self.cpc

    # TODO: Test for AdapterManager.__repr__()

    @pytest.mark.parametrize(
        "full_properties_kwargs, prop_names", [
            ({},
             ['object-uri', 'name', 'status']),
            (dict(full_properties=False),
             ['object-uri', 'name', 'status']),
            (dict(full_properties=True),
             None),
        ]
    )
    def test_adaptermanager_list_full_properties(
            self, full_properties_kwargs, prop_names):
        """Test AdapterManager.list() with full_properties."""

        # Add two faked adapters
        faked_osa1 = self.add_standard_osa()
        faked_hs2 = self.add_standard_hipersocket()

        exp_faked_adapters = [faked_osa1, faked_hs2]
        adapter_mgr = self.cpc.adapters

        # Execute the code to be tested
        adapters = adapter_mgr.list(**full_properties_kwargs)

        assert_resources(adapters, exp_faked_adapters, prop_names)

    @pytest.mark.parametrize(
        "list_kwargs, prop_names", [
            ({},
             ['object-uri', 'name', 'status']),
            (dict(additional_properties=[]),
             ['object-uri', 'name', 'status']),
            (dict(additional_properties=['description']),
             ['object-uri', 'name', 'status', 'description']),
            (dict(additional_properties=['description', 'detected-card-type']),
             ['object-uri', 'name', 'status', 'description',
              'detected-card-type']),
            (dict(additional_properties=['port-count']),
             ['object-uri', 'name', 'status', 'port-count']
             # port-count is not on every adapter
             ),
        ]
    )
    def test_adaptermanager_list_add_props(
            self, list_kwargs, prop_names):
        """Test AdapterManager.list() with additional_properties."""

        # Add two faked adapters
        faked_osa1 = self.add_standard_osa()
        faked_hs2 = self.add_standard_hipersocket()
        faked_cr3 = self.add_crypto_ce5s(self.faked_cpc)

        exp_faked_adapters = [faked_osa1, faked_hs2, faked_cr3]
        adapter_mgr = self.cpc.adapters

        # Execute the code to be tested
        adapters = adapter_mgr.list(**list_kwargs)

        assert_resources(adapters, exp_faked_adapters, prop_names)

    @pytest.mark.parametrize(
        "filter_args, exp_names", [
            ({'object-id': OSA1_OID},
             [OSA1_NAME]),
            ({'object-id': HS2_OID},
             [HS2_NAME]),
            ({'object-id': [OSA1_OID, HS2_OID]},
             [OSA1_NAME, HS2_NAME]),
            ({'object-id': [OSA1_OID, OSA1_OID]},
             [OSA1_NAME]),
            ({'object-id': OSA1_OID + 'foo'},
             []),
            ({'object-id': [OSA1_OID, HS2_OID + 'foo']},
             [OSA1_NAME]),
            ({'object-id': [HS2_OID + 'foo', OSA1_OID]},
             [OSA1_NAME]),
            ({'name': OSA1_NAME},
             [OSA1_NAME]),
            ({'name': HS2_NAME},
             [HS2_NAME]),
            ({'name': [OSA1_NAME, HS2_NAME]},
             [OSA1_NAME, HS2_NAME]),
            ({'name': OSA1_NAME + 'foo'},
             []),
            ({'name': [OSA1_NAME, HS2_NAME + 'foo']},
             [OSA1_NAME]),
            ({'name': [HS2_NAME + 'foo', OSA1_NAME]},
             [OSA1_NAME]),
            ({'name': [OSA1_NAME, OSA1_NAME]},
             [OSA1_NAME]),
            ({'name': '.*osa 1'},
             [OSA1_NAME]),
            ({'name': 'osa 1.*'},
             [OSA1_NAME]),
            ({'name': 'osa .'},
             [OSA1_NAME]),
            ({'name': '.sa 1'},
             [OSA1_NAME]),
            ({'name': '.+'},
             [OSA1_NAME, HS2_NAME]),
            ({'name': 'osa 1.+'},
             []),
            ({'name': '.+osa 1'},
             []),
            ({'name': OSA1_NAME,
              'object-id': OSA1_OID},
             [OSA1_NAME]),
            ({'name': OSA1_NAME,
              'object-id': OSA1_OID + 'foo'},
             []),
            ({'name': OSA1_NAME + 'foo',
              'object-id': OSA1_OID},
             []),
            ({'name': OSA1_NAME + 'foo',
              'object-id': OSA1_OID + 'foo'},
             []),
        ]
    )
    def test_adaptermanager_list_filter_args(self, filter_args, exp_names):
        """Test AdapterManager.list() with filter_args."""

        # Add two faked adapters
        self.add_standard_osa()
        self.add_standard_hipersocket()

        adapter_mgr = self.cpc.adapters

        # Execute the code to be tested
        adapters = adapter_mgr.list(filter_args=filter_args)

        assert len(adapters) == len(exp_names)
        if exp_names:
            names = [ad.properties['name'] for ad in adapters]
            assert set(names) == set(exp_names)

    def test_adaptermanager_create_hipersocket(self):
        """Test AdapterManager.create_hipersocket()."""

        hs_properties = {
            'name': 'hs 3',
            'description': 'Hipersocket #3',
            'port-description': 'Hipersocket #3 Port',
            'maximum-transmission-unit-size': 56,
        }

        adapter_mgr = self.cpc.adapters

        # Execute the code to be tested
        adapter = adapter_mgr.create_hipersocket(properties=hs_properties)

        assert isinstance(adapter, Adapter)
        assert adapter.name == adapter.properties['name']
        assert adapter.uri == adapter.properties['object-uri']

        # We expect the input properties to be in the resource object
        assert adapter.properties['name'] == 'hs 3'
        assert adapter.properties['description'] == 'Hipersocket #3'
        assert adapter.properties['port-description'] == 'Hipersocket #3 Port'
        assert adapter.properties['maximum-transmission-unit-size'] == 56

    # TODO: Test for initial Adapter attributes (ports, port_uris_prop,
    #       port_uri_segment)

    def test_adapter_repr(self):
        """Test Adapter.__repr__()."""

        # Add a faked adapter
        faked_osa = self.add_standard_osa()

        adapter_mgr = self.cpc.adapters
        adapter = adapter_mgr.find(name=faked_osa.name)

        # Execute the code to be tested
        repr_str = repr(adapter)

        repr_str = repr_str.replace('\n', '\\n')
        # We check just the begin of the string:
        assert re.match(
            rf'^{adapter.__class__.__name__}\s+at\s+'
            rf'0x{id(adapter):08x}\s+\(\\n.*',
            repr_str)

    def test_max_crypto_domains(self):
        """Test Adapter.maximum_crypto_domains on z13 and z13s."""

        faked_cpc = self.faked_cpc
        faked_crypto = self.add_crypto_ce5s(faked_cpc)
        self._one_test_max_crypto_domains(faked_cpc, faked_crypto, 85)

        faked_cpc = self.add_cpc_z13s()
        faked_crypto = self.add_crypto_ce5s(faked_cpc)
        self._one_test_max_crypto_domains(faked_cpc, faked_crypto, 40)

    def _one_test_max_crypto_domains(self, faked_cpc, faked_adapter,
                                     exp_max_domains):
        """Test Adapter.maximum_crypto_domains on a CPC"""

        cpc = self.client.cpcs.find(name=faked_cpc.name)
        adapter = cpc.adapters.find(name=faked_adapter.name)

        # Exercise code to be tested
        max_domains = adapter.maximum_crypto_domains

        assert max_domains == exp_max_domains

    def test_adapter_delete(self):
        """Test Adapter.delete() for Hipersocket adapter."""

        # Add two faked adapters
        self.add_standard_osa()
        faked_hs = self.add_standard_hipersocket()

        adapter_mgr = self.cpc.adapters
        hs_adapter = adapter_mgr.find(name=faked_hs.name)

        # Execute the code to be tested
        hs_adapter.delete()

        with pytest.raises(CeasedExistence):
            hs_adapter.pull_full_properties()

        with pytest.raises(NotFound):
            adapter_mgr.find(name=faked_hs.name)

    @pytest.mark.parametrize(
        "input_props", [
            {},
            {'description': 'New adapter description'},
            {'channel-path-id': '1A',
             'description': 'New adapter description'},
        ]
    )
    def test_adapter_update_properties(self, input_props):
        """Test Adapter.update_properties()."""

        # Add a faked adapter
        faked_adapter = self.add_standard_osa()

        adapter_mgr = self.cpc.adapters
        adapter = adapter_mgr.find(name=faked_adapter.name)

        adapter.pull_full_properties()
        saved_properties = copy.deepcopy(adapter.properties)

        # Execute the code to be tested
        adapter.update_properties(properties=input_props)

        # Verify that the resource object already reflects the property
        # updates.
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in adapter.properties
            prop_value = adapter.properties[prop_name]
            assert prop_value == exp_prop_value

        # Refresh the resource object and verify that the resource object
        # still reflects the property updates.
        adapter.pull_full_properties()
        for prop_name in saved_properties:
            if prop_name in input_props:
                exp_prop_value = input_props[prop_name]
            else:
                exp_prop_value = saved_properties[prop_name]
            assert prop_name in adapter.properties
            prop_value = adapter.properties[prop_name]
            assert prop_value == exp_prop_value

    def test_adapter_update_name(self):
        """
        Test Adapter.update_properties() with 'name' property.
        """

        # Add a faked adapter
        faked_adapter = self.add_standard_osa()
        adapter_name = faked_adapter.name

        adapter_mgr = self.cpc.adapters
        adapter = adapter_mgr.find(name=adapter_name)

        new_adapter_name = "new-" + adapter_name

        # Execute the code to be tested
        adapter.update_properties(properties={'name': new_adapter_name})

        # Verify that the resource is no longer found by its old name, using
        # list() (this does not use the name-to-URI cache).
        adapters_list = adapter_mgr.list(filter_args=dict(name=adapter_name))
        assert len(adapters_list) == 0

        # Verify that the resource is no longer found by its old name, using
        # find() (this uses the name-to-URI cache).
        with pytest.raises(NotFound):
            adapter_mgr.find(name=adapter_name)

        # Verify that the resource object already reflects the update, even
        # though it has not been refreshed yet.
        assert adapter.properties['name'] == new_adapter_name

        # Refresh the resource object and verify that it still reflects the
        # update.
        adapter.pull_full_properties()
        assert adapter.properties['name'] == new_adapter_name

        # Verify that the resource can be found by its new name, using find()
        new_adapter_find = adapter_mgr.find(name=new_adapter_name)
        assert new_adapter_find.properties['name'] == new_adapter_name

        # Verify that the resource can be found by its new name, using list()
        new_adapters_list = adapter_mgr.list(
            filter_args=dict(name=new_adapter_name))
        assert len(new_adapters_list) == 1
        new_adapter_list = new_adapters_list[0]
        assert new_adapter_list.properties['name'] == new_adapter_name

    # TODO: Test for Adapter.change_crypto_type()

    @pytest.mark.parametrize(
        "init_type", ['not-configured', 'fc', 'fcp']
    )
    @pytest.mark.parametrize(
        "new_type", ['not-configured', 'fc', 'fcp']
    )
    def test_change_adapter_type_success(self, init_type, new_type):
        """Test Adapter.change_adapter_type() on ficon adapter with success."""

        faked_cpc = self.faked_cpc
        faked_adapter = self.add_ficon_fe6sp(faked_cpc)

        # Set the desired initial adapter type for the test
        faked_adapter.properties['type'] = init_type

        adapter_mgr = self.cpc.adapters
        adapter = adapter_mgr.find(name=faked_adapter.name)

        if new_type == init_type:
            with pytest.raises(HTTPError) as exc_info:

                # Execute the code to be tested
                adapter.change_adapter_type(new_type)

            exc = exc_info.value
            assert exc.http_status == 400
            assert exc.reason == 8
        else:

            # Execute the code to be tested.
            adapter.change_adapter_type(new_type)

            act_type = adapter.get_property('type')
            assert act_type == new_type

    @pytest.mark.parametrize(
        "desc, family, init_type, new_type, exp_exc", [
            (
                "Invalid adapter family: 'osa'",
                'osa', 'osd', None,
                HTTPError({'http-status': 400, 'reason': 18})
            ),
            (
                "Invalid new type value: 'xxx'",
                'ficon', 'fcp', 'xxx',
                HTTPError({'http-status': 400, 'reason': 8})
            ),
        ]
    )
    def test_change_adapter_type_error(
            self, desc, family, init_type, new_type, exp_exc):
        # pylint: disable=unused-argument
        """Test Adapter.change_adapter_type()."""

        faked_cpc = self.faked_cpc
        if family == 'ficon':
            faked_adapter = self.add_ficon_fe6sp(faked_cpc)
        else:
            assert family == 'osa'
            faked_adapter = self.add_standard_osa()

        faked_adapter.properties['type'] = init_type

        adapter_mgr = self.cpc.adapters
        adapter = adapter_mgr.find(name=faked_adapter.name)

        with pytest.raises(exp_exc.__class__) as exc_info:

            # Execute the code to be tested
            adapter.change_adapter_type(new_type)

        exc = exc_info.value
        if isinstance(exp_exc, HTTPError):
            assert exc.http_status == exp_exc.http_status
            assert exc.reason == exp_exc.reason



# Copyright 2024 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The :class:`zhmcclient.OSConsole` class encapsulates the interactions with
the console of an operating system running in a partition, by using the
WebSocket protocol supported by the HMC, which connects that internally to the
OS console.

The OS console can also be interacted with through any WebSocket client,
using the WebSocket URI returned by the
:meth:`zhmcclient.Partition.create_os_websocket` method. Thus, the use of this
class is not mandatory but it simplifies some of the interactions quite
significantly.

Example code using the class to login and execute a 'uname -a' command:

.. code-block:: python

    partition = ...  # a zhmcclient.Partition object

    cons = zhmcclient.OSConsole(partition)

    try:
        cons.connect()
        cons.login_linux('admin', 'password')

        output = cons.execute_command('uname -a')
        print(f"uname -a: {output}")

    finally:
        if cons.is_connected():
            cons.disconnect()
"""

import re
import ssl
import websocket
import certifi

from ._constants import DEFAULT_HMC_PORT, DEFAULT_WS_TIMEOUT, OS_LOGGER_NAME
from ._exceptions import OSConsoleConnectedError, OSConsoleNotConnectedError, \
    OSConsoleWebSocketError, OSConsoleAuthError
from ._logging import get_logger

__all__ = ['OSConsole']

OS_LOGGER = get_logger(OS_LOGGER_NAME)


class OSConsole:
    """
    This class encapsulates the interactions with the console of an operating
    system running in a partition, by using the WebSocket protocol supported by
    the HMC.

    HMC/SE version requirements:

    * HMC version >= 2.14.0 with HMC API version >= 2.22
    """

    def __init__(self, partition, ws_timeout=DEFAULT_WS_TIMEOUT):
        """
        Parameters:

          partition (:class:`~zhmcclient.Partition`): Targeted partition for
            connecting to its operating system console.

          ws_timeout (int): WebSocket connect and read timeout in seconds.
            This timeout is used to detect the end of output on the OS console.
            If it is too short, output will be missed. The timeout should not
            be shorter than 5 seconds.
            Default: :attr:`~zhmcclient._constants.DEFAULT_WS_TIMEOUT`.
        """
        self.partition = partition
        self.ws_timeout = ws_timeout

        self.ws_uri = None
        self.ws_full_uri = None
        self.ws = None
        self.previous_text = None

    def _check_connected(self):
        if self.ws is None:
            raise OSConsoleNotConnectedError(
                "Not connected to the OS console")

    def _check_not_connected(self):
        if self.ws is not None:
            cpc = self.partition.manager.parent
            raise OSConsoleConnectedError(
                "Already connected to the OS console for partition "
                f"{cpc.name}.{self.partition.name}")

    def is_connected(self):
        """
        Return a boolean indicating whether this object is currently connected
        to the OS console.
        """
        return self.ws is not None

    def connect(self):
        """
        Create a WebSocket on the HMC and use it to connect to the OS console.

        This object must not currently be connected to the OS console.

        Raises:

          zhmcclient.Error: Issue when creating the WebSocket on the HMC.
          OSConsoleConnectedError: Already connected to the OS console.
          OSConsoleWebSocketError: Issue when connecting via the WebSocket.
        """
        self._check_not_connected()

        self.ws_uri = self.partition.create_os_websocket()
        # may raise zhmcclient.Error

        session = self.partition.manager.session
        self.ws_full_uri = \
            f"wss://{session.actual_host}:{DEFAULT_HMC_PORT}{self.ws_uri}"

        sslopt = {}
        if isinstance(session.verify_cert, str):
            sslopt["cert_reqs"] = ssl.CERT_REQUIRED
            sslopt["ca_cert_path"] = session.verify_cert
        elif session.verify_cert is True:
            sslopt["cert_reqs"] = ssl.CERT_REQUIRED
            sslopt["ca_cert_path"] = certifi.where()
        else:
            sslopt["cert_reqs"] = ssl.CERT_NONE
        self.ws = websocket.WebSocket(sslopt=sslopt)

        try:
            self.ws.connect(self.ws_full_uri, timeout=self.ws_timeout)
        except (websocket.WebSocketException, OSError) as exc:
            raise OSConsoleWebSocketError(str(exc))

        self.previous_text = None

    def disconnect(self):
        """
        Disconnect from the OS console.

        This object must currently be connected to the OS console.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when closing the WebSocket.
        """
        self._check_connected()

        try:
            self.ws.close()
        except (websocket.WebSocketException, OSError) as exc:
            raise OSConsoleWebSocketError(str(exc))

        self.ws = None
        self.ws_uri = None
        self.ws_full_uri = None
        self.previous_text = None

    def recv_all(self):
        """
        Receive all text data from the OS console until no more data is
        available within the WebSocket timeout, and return that.

        This object must currently be connected to the OS console.

        In the string that is returned, any sequence of one or more CR (U+000D)
        characters followed by NL (U+000A) has been translated to NL (i.e. the
        CR characters are removed).

        Returns:

          str: Received text data, as a Unicode string.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when reading data from the WebSocket.
        """
        self._check_connected()

        data_items = []
        if self.previous_text is not None:
            data_items.append(self.previous_text)
            self.previous_text = None
        while True:
            try:
                data = self.ws.recv()
            except websocket.WebSocketTimeoutException:
                return ''.join(data_items)
            except (websocket.WebSocketException, OSError) as exc:
                raise OSConsoleWebSocketError(str(exc))

            assert isinstance(data, str), f"data is not a string: {data!r}"
            data = re.sub(r'\r+\n', r'\n', data)
            data_items.append(data)

    def recv_line(self):
        """
        Receive one line of text data from the OS console and return that.

        This object must currently be connected to the OS console.

        This may include possibly multiple calls to receive messages from
        the WebSocket. Any excess data is buffered and is processed by
        the next call to :meth:`recv_line` or :meth:`recv_all`.

        A line end is detected in the data as follows:

        * when an NL (U+000A) character is encountered
        * when the WebSocket read timeout has expired without encountering NL

        In the string that is returned, any sequence of one or more CR (U+000D)
        characters followed by NL (U+000A) has been translated to NL (i.e. the
        CR characters are removed).

        Returns:

          str: The line of text data, normally terminated by NL. If the
          WebSocket read timeout has expired without encountering NL, the
          string will not have a trailing NL. If the WebSocket read timeout
          has expired without getting any data, the empty string will be
          returned.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when reading data from the WebSocket.
        """
        self._check_connected()

        line_parts = []
        while True:

            # Get data if there is no previous data
            if not self.previous_text:
                try:
                    text = self.ws.recv()
                except websocket.WebSocketTimeoutException:
                    break
                except (websocket.WebSocketException, OSError) as exc:
                    raise OSConsoleWebSocketError(str(exc))
                # The HMC always returns text data:
                assert isinstance(text, str)
                self.previous_text = re.sub(r'\r+\n', r'\n', text)

            # Process the data to extract the line
            line, end, remainder = self.previous_text.partition('\n')
            line_parts.append(line)
            line_parts.append(end)
            self.previous_text = remainder
            if end:
                break

        return ''.join(line_parts)

    def send(self, text):
        """
        Send text data to the OS console.

        This object must currently be connected to the OS console.

        The data can be one or more lines, separated by NL, and typically also
        terminated by NL. The lines are interpreted as commands in the OS
        console.

        Parameters:

          text (str): The text data, as a Unicode string.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when sending data to the WebSocket.
        """
        self._check_connected()

        self.ws.send_text(text)

    def login_linux(self, username, password):
        """
        Login to a Linux operating system running in the partition.

        This object must currently be connected to the OS console.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when sending data to or receiving
            data from the WebSocket.
          OSConsoleAuthError: Authentication issue when logging in to Linux.
        """
        # The check for connection is performed in self.recv_all()
        while True:
            lines = self.recv_all()
            OS_LOGGER.debug("login_linux: Received: %r", lines)
            if lines and re.search(
                    r"login:$", lines.strip(), flags=re.I + re.M):
                OS_LOGGER.debug(
                    "login_linux: Found login prompt, sending username")
                self.send(username + '\n')
                lines = self.recv_all()
                OS_LOGGER.debug("Received: %r", lines)
                if lines and re.search(
                        r"password:$", lines.strip(), flags=re.I + re.M):
                    OS_LOGGER.debug(
                        "login_linux: Found password prompt, sending password")
                    self.send(password + '\n')
                    lines = self.recv_all()
                    OS_LOGGER.debug("login_linux: Received: %r", lines)
                    if lines and re.search(
                            r"(login incorrect|login timed out)",
                            lines, flags=re.I + re.M):
                        msg = re.sub(
                            r"\n[^\n]*login:", "", lines, flags=re.I + re.M)
                        msg = msg.strip().replace('\n', ' ')
                        raise OSConsoleAuthError(msg)
                    OS_LOGGER.debug("login_linux: Successfully logged in")
                    break
            else:
                OS_LOGGER.debug(
                    "login_linux: Sending empty line to get to login prompt")
                self.send('\n')

    def execute_command(self, command):
        """
        Execute a command in the OS console and return its output lines.

        This object must currently be connected to the OS console.
        Depending on the operating system, this object also needs to be logged
        in to the OS.

        The execution environment for the commands is always the console of
        the OS, so it depends on the type of OS:

        * On z/VM, the commands are executed as CP commands.
          The command output is returned as output lines.

        * On Linux, the commands are executed as normal shell commands
          in the console environment.
          The command output with stdout and stderr merged into one stream is
          returned as output lines.

        The exit code of the commands is not available.

        In the output lines string that is returned, any sequence of one or
        more CR (U+000D) characters followed by NL (U+000A) has been translated
        to NL (i.e. the CR characters are removed).

        Parameters:

          command (str): Command string (without any trailing NL).

        Returns:

          str: Command output lines. The lines are separated by NL.

        Raises:

          OSConsoleNotConnectedError: Not connected to the OS console.
          OSConsoleWebSocketError: Issue when sending data to or receiving
            data from the WebSocket.
        """
        # The check for connection is performed in self.send()
        OS_LOGGER.debug("execute_command: Sending command: %r", command)
        self.send(command + '\n')
        output = self.recv_all()
        OS_LOGGER.debug("execute_command: Received command output: %r", output)
        return output



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for end2end tests.
"""


import os
import re
import random
import time
import warnings
import logging
import pytest

import zhmcclient

# Prefix used for names of resources that are created during tests
TEST_PREFIX = 'zhmcclient_tests_end2end'


def setup_logging(enable_logging, testfunc_name, log_file=None):
    """
    Set up logging for end2end testcases.

    If enable_logging is True, two loggers are created and enabled for logging
    to the specified log_file:
      * A logger named "zhmcclient.hmc" for logging the HMC interactions.
        That logger is used by the zhmcclient package.
      * A logger named testfunc_name for logging additional information the
        test function may need to log. This logger is returned.
    Because this setup function is called for each invocation of a test
    function, it ends up being called multiple times within the same Python
    process. Therefore, the loggers are set up only when they do not exist yet.

    If enable_logging is False, the same two loggers are created and logging
    is disabled on them by setting the log level to NOTSET. The testfunc_name
    logger is also returned in this case.

    Returns:

        logging.Logger: Logger for testfunc_name
    """
    # log level and format for both loggers
    log_level = logging.DEBUG
    log_format = '%(asctime)s %(levelname)s %(name)s: %(message)s'
    datetime_format = '%Y-%m-%d %H:%M:%S %Z'
    log_converter = time.gmtime

    zhmcclient_logger = logging.getLogger(zhmcclient.HMC_LOGGER_NAME)
    testfunc_logger = logging.getLogger(testfunc_name)

    if enable_logging:
        assert log_file

        abs_log_file = os.path.abspath(log_file)

        zhmcclient_handler = None
        for handler in zhmcclient_logger.handlers:
            if isinstance(handler, logging.FileHandler):
                bfn = handler.baseFilename  # pylint: disable=no-member
                abs_handler_file = os.path.abspath(bfn)
                if abs_handler_file == abs_log_file:
                    zhmcclient_handler = handler
                    break
        if not zhmcclient_handler:
            zhmcclient_handler = logging.FileHandler(log_file)
            zhmcclient_logger.addHandler(zhmcclient_handler)

        testfunc_handler = None
        for handler in testfunc_logger.handlers:
            if isinstance(handler, logging.FileHandler):
                bfn = handler.baseFilename  # pylint: disable=no-member
                abs_handler_file = os.path.abspath(bfn)
                if abs_handler_file == abs_log_file:
                    testfunc_handler = handler
                    break
        if not testfunc_handler:
            testfunc_handler = logging.FileHandler(log_file)
            testfunc_logger.addHandler(testfunc_handler)

        logging.Formatter.converter = log_converter
        formatter = logging.Formatter(
            log_format, datefmt=datetime_format)

        zhmcclient_handler.setFormatter(formatter)
        testfunc_handler.setFormatter(formatter)

        zhmcclient_logger.setLevel(log_level)
        testfunc_logger.setLevel(log_level)

    else:
        zhmcclient_logger.setLevel(logging.NOTSET)
        testfunc_logger.setLevel(logging.NOTSET)

    return testfunc_logger


class End2endTestWarning(UserWarning):
    """
    Python warning indicating an issue with an end2end test.
    """
    pass


def assert_res_props(res, exp_props, ignore_values=None, prop_names=None):
    """
    Check the properties of a resource object.
    """
    res_props = dict(res.properties)
    # checked_prop_names = set()
    for prop_name in exp_props:

        if prop_names is not None and prop_name not in prop_names:
            continue  # Only check properties in prop_names

        assert prop_name in res_props, (
            f"Property {prop_name!r} not found in {res.prop('class')} "
            f"object {res.name!r}")

        if ignore_values is not None and prop_name not in ignore_values:
            act_value = res_props[prop_name]
            exp_value = exp_props[prop_name]
            assert_res_prop(act_value, exp_value, prop_name, res)

        # checked_prop_names.add(prop_name)

    # extra_prop_names = set(res_props.keys()) - checked_prop_names

    # TODO: Decide whether we want to check the exact set, or the minimum set.
    # assert not extra_prop_names, (
    #     "The following properties were unexpectedly present in "
    #     f"{res.prop('class')} object{res.name!r} : "
    #     f"{', '.join(extra_prop_names)}")


def assert_res_prop(act_value, exp_value, prop_name, res):
    """
    Check a property of a resource object.
    """
    assert act_value == exp_value, (
        f"Property {prop_name!r} has unexpected value in {res.prop('class')} "
        f"object {res.name!r}: Expected: {exp_value!r}, actual: {act_value!r}")


def _res_name(item):
    """Return the resource name, used by pick_test_resources()"""
    if isinstance(item, (tuple, list)):
        return item[0].name
    return item.name


def pick_test_resources(res_list):
    """
    Return the list of resources to be tested.

    The env.var "TESTRESOURCES" controls which resources are picked for the
    test, as follows:

    * 'random': (default) one random choice from the input list of resources.
    * 'all': the complete input list of resources.
    * '<pattern>': The resources with names matching the regexp pattern.

    Parameters:

      res_list (list of zhmcclient.BaseResource or tuple thereof):
        List of resources to pick from. Tuple items are a resource and its
        parent resources.

    Returns:

      list of zhmcclient.BaseResource: Picked list of resources.
    """

    test_res = os.getenv('TESTRESOURCES', 'random')

    if test_res == 'random':
        return [random.choice(res_list)]

    if test_res == 'all':
        return sorted(res_list, key=_res_name)

    # match the pattern
    ret_list = []
    for item in res_list:
        if re.search(test_res, _res_name(item)):
            ret_list.append(item)
    return sorted(ret_list, key=_res_name)


def runtest_find_list(session, manager, name, server_prop, client_prop,
                      volatile_props, minimal_props, list_props, add_props=None,
                      unique_name=True):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Run tests for find/list methods for a resource type:
    - find_by_name(name) (only if not unique_name)
    - pull_full_properties()
    - find(**filter_args)
      - server-side filter
    - findall(**filter_args)
      - no filter
      - server-side filter
      - client-side filter
    - list(full_properties=False, filter_args=None, additional_properties=None)
      - no filter + short
      - no filter + additional_properties
      - server-side filter + full
      - server-side filter + short
      - client-side filter + short

    Parameters:

      session (zhmcclient.Session): Session to an HMC.

      manager (zhmcclient.BaseManager): Manager for listing the resources.

      name (string): Name of resource to be found.

      server_prop (string): Name of resource property for server-side filtering,
        or None for no tests with server-side filtering.

      client_prop (string): Name of resource property for client-side filtering,
        or None for no tests with client-side filtering.

      volatile_props (list of string): Names of properties that are ignored when
        comparing resources because they are volatile.

      minimal_props (list of string): Names of properties that are in
        minimalistic resources (e.g. returned by find_by_name()).

      list_props (list of string): Names of properties that are returned by
        list()).

      add_props (list of string): Names of additional properties to be returned
        by list(additional_properties)). If None, the additional properties test
        is not executed.

      unique_name (bool): Indicates that the resource name is expected to be
        unique within its parent resource. That is normally the case, the only
        known exception are storage volumes in HMC 2.14.0 (they were introduced
        in 2.14.0 and made unique in 2.14.1).
    """
    if unique_name:
        # The code to be tested: find_by_name(name)
        found_res = manager.find_by_name(name)

        # Get full properties directly, for comparison
        exp_props = session.get(found_res.uri)

        if client_prop:
            client_value = exp_props[client_prop]

        if server_prop:
            server_value = exp_props[server_prop]

        # Get the object-id of the resource from its URI
        oid = found_res.uri.split('/')[-1]
        oid_prop = manager._oid_prop  # pylint: disable=protected-access

        assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                         prop_names=minimal_props)

        # The code to be tested: find(oid), for optimized lookup
        found_res = manager.find(**{oid_prop: oid})

        assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                         prop_names=minimal_props)

        # The code to be tested: pull_full_properties()
        found_res.pull_full_properties()

        assert_res_props(found_res, exp_props, ignore_values=volatile_props)

        if server_prop:
            # The code to be tested: find() with server-side filter
            found_res = manager.find(**{server_prop: server_value})

            assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                             prop_names=minimal_props)

            # The code to be tested: findall() with server-side filter
            found_res_list = manager.findall(**{server_prop: server_value})

            assert len(found_res_list) == 1
            found_res = found_res_list[0]
            assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                             prop_names=minimal_props)

        if client_prop:
            # The code to be tested: findall() with client-side filter
            # pylint: disable=used-before-assignment
            found_res_list = manager.findall(**{client_prop: client_value})

            assert name in map(lambda _res: _res.name, found_res_list)
            found_res_list = list(filter(lambda _res: _res.name == name,
                                         found_res_list))
            found_res = found_res_list[0]
            if len(found_res_list) > 1:
                raise AssertionError(
                    f"{found_res.prop('class')} findall(client_filter) result "
                    f"with non-unique name {name!r}: {found_res_list}")
            assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                             prop_names=list_props)

        if server_prop:
            # Code to be tested: list() with server-side filter and full props
            found_res_list = manager.list(
                full_properties=True, filter_args={server_prop: server_value})

            assert len(found_res_list) == 1
            found_res = found_res_list[0]
            assert_res_props(found_res, exp_props, ignore_values=volatile_props)

        if server_prop:
            # Code to be tested: list() with server-side filter and short props
            found_res_list = manager.list(
                filter_args={server_prop: server_value})

            assert len(found_res_list) == 1
            found_res = found_res_list[0]
            assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                             prop_names=minimal_props)

        if client_prop:
            # Code to be tested: list() with client-side filter and short props
            found_res_list = manager.list(
                filter_args={client_prop: client_value})

            assert name in [_res.name for _res in found_res_list]
            found_res = [_res for _res in found_res_list
                         if _res.name == name][0]
            assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                             prop_names=minimal_props)

    # The code to be tested: findall() with no filter
    found_res_list = manager.findall()

    assert name in map(lambda _res: _res.name, found_res_list)
    found_res_list = list(filter(lambda _res: _res.name == name,
                                 found_res_list))
    found_res = found_res_list[0]
    if len(found_res_list) > 1:
        found_uri_list = [r.uri for r in found_res_list]
        parent = manager.parent
        uri_str = '\n'.join(found_uri_list)
        raise AssertionError(
            f"{found_res.prop('class')} findall() result for "
            f"{parent.prop('class')} {parent.name!r} has non-unique name "
            f"{name!r} for the following {len(found_res_list)} objects:\n"
            f"{uri_str}")
    assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                     prop_names=list_props)

    # The code to be tested: list() with no filter and short properties
    found_res_list = manager.list()

    # Because we have only the expected properties for the resource with
    # the specified name, we only compare that one resource.
    assert name in map(lambda _res: _res.name, found_res_list)
    found_res_list = list(filter(lambda _res: _res.name == name,
                                 found_res_list))
    found_res = found_res_list[0]
    if len(found_res_list) > 1:
        raise AssertionError(
            f"{found_res.prop('class')} list() result with non-unique name "
            f"{name!r}: {found_res_list}")
    assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                     prop_names=list_props)

    if add_props is not None:
        # The code to be tested: list() with no filter and additional properties
        found_res_list = manager.list(additional_properties=add_props)

        # Because we have only the expected properties for the resource with
        # the specified name, we only compare that one resource.
        assert name in map(lambda _res: _res.name, found_res_list)
        found_res_list = list(filter(lambda _res: _res.name == name,
                                     found_res_list))
        found_res = found_res_list[0]
        if len(found_res_list) > 1:
            raise AssertionError(
                f"{found_res.prop('class')} list() result with non-unique "
                f"name {name!r}: {found_res_list}")
        assert_res_props(found_res, exp_props, ignore_values=volatile_props,
                         prop_names=list_props + add_props)


def runtest_get_properties(manager, non_list_prop):  # noqa: F811
    # pylint: disable=redefined-outer-name
    """
    Run tests for pull_full_properties/pull_properties/get_property/prop
    methods for a resource type.

    Parameters:

      manager (zhmcclient.BaseManager): Manager for listing the resources.

      non_list_prop (string): Name of a resource property that is not returned
        with the List HMC operation.
    """

    assert isinstance(manager, zhmcclient.BaseManager)

    # Indicates whether pull_properties() is expected to pull just the
    # specified properties. We make this dependent only on whether the
    # mock support for the resource implements it, and not on the HMC
    # version, because the zhmcclient mock support only implements the
    # latest CPC and HMC generation, for simplicity.
    supports_properties = manager.supports_properties

    # Part 1: First chunk of methods to be tested

    # Get a fresh resource with list properties
    manager.invalidate_cache()
    resources = manager.list()
    resource = random.choice(resources)

    local_pnames = set(resource.properties.keys())
    local_pnames_plus = local_pnames | {non_list_prop}

    # Validate that the non_list_prop property is not listed.
    # This is really just checking that the testcase was invoked correctly.
    assert non_list_prop not in local_pnames, (
        f"non_list_prop={non_list_prop!r}, local_pnames={local_pnames!r}")

    # Validate initial state of the resource w.r.t. properties
    assert resource.full_properties is False

    # Validate that get_property() does not pull additional properties
    for pname in local_pnames:
        _ = resource.get_property(pname)
        assert resource.full_properties is False, \
            f"resource={resource!r}"
        current_pnames = set(resource.properties.keys())
        assert current_pnames == local_pnames, (
            f"current_pnames={current_pnames!r}, local_pnames={local_pnames!r}")

    # Validate that prop() does not pull additional properties
    for pname in local_pnames:
        _ = resource.prop(pname)
        assert resource.full_properties is False, \
            f"resource={resource!r}"
        current_pnames = set(resource.properties.keys())
        assert current_pnames == local_pnames, (
            f"current_pnames={current_pnames!r}, local_pnames={local_pnames!r}")

    # Validate that get_property() on non_list_prop pulls full properties
    _ = resource.get_property(non_list_prop)
    assert resource.full_properties is True, \
        f"resource={resource!r}"
    current_pnames = set(resource.properties.keys())
    assert current_pnames > local_pnames_plus, (
        f"current_pnames={current_pnames!r}, "
        f"local_pnames_plus={local_pnames_plus!r}")

    # Part 2: Second chunk of methods to be tested

    # Get a fresh resource with list properties
    manager.invalidate_cache()
    resources = manager.list()
    resource = random.choice(resources)

    local_pnames = set(resource.properties.keys())
    local_pnames_plus = local_pnames | {non_list_prop}

    # Validate initial state of the resource w.r.t. properties
    assert resource.full_properties is False

    # Validate that pull_properties() with no properties does not fetch props
    if supports_properties:
        prior_pnames = set(resource.properties.keys())
        resource.pull_properties([])
        current_pnames = set(resource.properties.keys())
        assert current_pnames == prior_pnames, (
            f"current_pnames={current_pnames!r}, "
            f"prior_pnames={prior_pnames!r}")

    # Validate pull_properties() with a non-list-result property
    resource.pull_properties([non_list_prop])
    current_pnames = set(resource.properties.keys())
    if supports_properties:
        # We get just the specified property in addition
        assert current_pnames == local_pnames_plus, (
            f"current_pnames={current_pnames!r}, "
            f"local_pnames_plus={local_pnames_plus!r}")
        assert resource.full_properties is False, \
            f"resource={resource!r}"
    else:
        # We get the full set of properties
        assert resource.full_properties is True, \
            f"resource={resource!r}"

    # Part 3: Third chunk of methods to be tested

    # Get a fresh resource with list properties
    manager.invalidate_cache()
    resources = manager.list()
    resource = random.choice(resources)

    local_pnames = set(resource.properties.keys())
    assert resource.full_properties is False

    # Validate that pull_full_properties() pulls full properties
    resource.pull_full_properties()
    assert resource.full_properties is True
    current_pnames = set(resource.properties.keys())
    assert current_pnames > local_pnames_plus, (
        f"current_pnames={current_pnames!r}, "
        f"local_pnames_plus={local_pnames_plus!r}")

    # Validate that pull_properties() with non_list_prop still has full props
    resource.pull_properties([non_list_prop])
    assert resource.full_properties is True


def validate_list_features(api_version, all_features, regex_reduced_features):
    """
    Tests the (already retrieved) results of calls to list_api_calls().
    Can be used for validating the corresponding method results on Console and
    CPC objects.

    Parameters:

        api_version: the result of a call to client.query_api_version()

        all_features: the result of a call to  list_api_calls()

        regex_reduced_features: the result of a call to  list_api_calls('cpc.*')
    """
    assert len(regex_reduced_features) <= len(all_features)

    wsapi_version = (api_version['api-major-version'],
                     api_version['api-minor-version'])

    if wsapi_version < (4, 10):
        # API features aren't supported prior 4.10, list must be empty
        assert len(all_features) == 0
        return

    # Even when API features are supported, the lists can still be empty.
    # (for example when HMC/SE driver wasn't restarted after features
    # where enabled)
    if len(all_features) > 0:
        # But when there are some available API features, there are a few that
        # are always present.
        expected_features = ['cpc-install-and-activate']
        for feature in expected_features:
            assert feature in all_features, \
                f'{feature} missing from {all_features}'
            assert feature in regex_reduced_features, \
                f'{feature} missing from {regex_reduced_features}'

        # Ensure pattern matching using 'cpc.*' worked
        assert len(regex_reduced_features) < len(all_features)
        assert 'report-a-problem' in all_features
        assert 'report-a-problem' not in regex_reduced_features


def skipif_no_storage_mgmt_feature(cpc):
    """
    Skip the test if the "DPM Storage Management" feature is not enabled for
    the specified CPC, or if the CPC does not yet support it.
    """
    try:
        smf = cpc.feature_enabled('dpm-storage-management')
    except ValueError:
        smf = False
    if not smf:
        skip_warn("DPM Storage Mgmt feature not enabled or not supported "
                  f"on CPC {cpc.name}")


def skipif_storage_mgmt_feature(cpc):
    """
    Skip the test if the "DPM Storage Management" feature is enabled for
    the specified CPC.
    """
    try:
        smf = cpc.feature_enabled('dpm-storage-management')
    except ValueError:
        smf = False
    if smf:
        skip_warn(f"DPM Storage Mgmt feature enabled on CPC {cpc.name}")


def skipif_no_group_support(client):
    """
    Skip the test if the HMC version does not support groups yet.
    """
    api_version = client.query_api_version()
    hmc_version = api_version['hmc-version']
    hmc_version_info = tuple(map(int, hmc_version.split('.')))
    if hmc_version_info < (2, 13, 0):
        skip_warn(
            f"HMC has version {hmc_version} and does not yet support groups")


def skipif_no_secure_boot_feature(cpc):
    """
    Skip the test if the API feature "secure-boot-with-certificates" isn't
    available on the specified CPC & console.
    """
    _skipif_api_feature_not_on_cpc_and_hmc("secure-boot-with-certificates", cpc)


def _skipif_api_feature_not_on_cpc_and_hmc(feature, cpc):
    """
    Skip the test if the given API feature isn't available on the specified CPC
     and its console.
    """
    cpc_features = cpc.list_api_features()

    if feature not in cpc_features:
        skip_warn(f"API feature {feature} not available on CPC {cpc.name}")

    console = cpc.manager.client.consoles.console
    console_features = console.list_api_features()
    if feature not in console_features:
        skip_warn(f"API feature {feature} not available on HMC {console.name}")


def has_api_feature(feature, cpc):
    """
    Returns True if the given API feature is available on the specified CPC
    and its console.
    """
    cpc_features = cpc.list_api_features()
    console = cpc.manager.client.consoles.console
    console_features = console.list_api_features()

    return feature in cpc_features and feature in console_features


def standard_partition_props(cpc, part_name):
    """
    Return the input properties for a standard partition in the specified CPC.
    """
    part_input_props = {
        'name': part_name,
        'description': 'Test partition for zhmcclient end2end tests',
        'initial-memory': 1024,
        'maximum-memory': 2048,
        'processor-mode': 'shared',  # used for filtering
        'type': 'linux',  # used for filtering
    }
    if cpc.prop('processor-count-ifl'):
        part_input_props['ifl-processors'] = 2
    elif cpc.prop('processor-count-general-purpose'):
        part_input_props['cp-processors'] = 2
    else:
        part_input_props['cp-processors'] = 1
        pc_names = filter(lambda p: p.startswith('processor-count-'),
                          cpc.properties.keys())
        pc_list = [f"{n}={cpc.properties[n]}" for n in pc_names]
        warnings.warn(
            f"CPC {cpc.name} shows neither IFL nor CP processors, "
            "specifying 1 CP for partition creation. "
            f"CPC processor-count properties are: {', '.join(pc_list)}",
            End2endTestWarning)

    return part_input_props


def skip_warn(msg):
    """
    Issue an End2endTestWarning and skip the current pytest testcase with the
    specified message.
    """
    warnings.warn(msg, End2endTestWarning, stacklevel=2)
    pytest.skip(msg)


def cleanup_and_import_example_certificate(cpc):
    """
    Removes any existing example certificate, then imports an example
    certificate.

    The certificate is valid until 2033-04-15, but should probably be regularly
    renewed.

    Returns the imported Certificate objects and the dict used for import
    """
    console = cpc.manager.console

    cert_name = f"{TEST_PREFIX} timestamp {time.strftime('%H.%M.%S')}"
    cert_name_new = cert_name + ' updated'

    try:
        cert = console.certificates.find(name=cert_name)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test cert from previous run: {cert_name!r} on CPC "
            f"{cpc.name}", UserWarning)
        cert.delete()
    try:
        cert = console.certificates.find(name=cert_name_new)
    except zhmcclient.NotFound:
        pass
    else:
        warnings.warn(
            f"Deleting test cert from previous run: {cert_name_new!r} on "
            f"CPC {cpc.name}", UserWarning)
        cert.delete()
    props = {
        # pylint: disable=line-too-long
        "certificate": "MIIFdjCCA14CCQCILyUhzc9RUjANBgkqhkiG9w0BAQsFADB9MQswCQYDVQQGEwJVUzELMAkGA1UECAwCTlkxDzANBgNVBAcMBkFybW9uazETMBEGA1UECgwKemhtY2NsaWVudDETMBEGA1UECwwKemhtY2NsaWVudDEmMCQGA1UEAwwdaHR0cHM6Ly9naXRodWIuY29tL3pobWNjbGllbnQwHhcNMjMwNDE4MDY0OTAyWhcNMzMwNDE1MDY0OTAyWjB9MQswCQYDVQQGEwJVUzELMAkGA1UECAwCTlkxDzANBgNVBAcMBkFybW9uazETMBEGA1UECgwKemhtY2NsaWVudDETMBEGA1UECwwKemhtY2NsaWVudDEmMCQGA1UEAwwdaHR0cHM6Ly9naXRodWIuY29tL3pobWNjbGllbnQwggIiMA0GCSqGSIb3DQEBAQUAA4ICDwAwggIKAoICAQDH8F7xMisAESV35LqjC3p6AsrGZ3kEOE5W8wcy0q7TEF9TO7TsAPdWit0e7WT20R1gYErv9uyFJeI4idIjWgTT8GVPixwcXClyywh/ND54voHrMZdbDGbvs5+wcfX/7BDtjRUtzuMtvDswEZqaQU/2W+rDRpb/FolwXDTNm17dSomegm7sw8xQsZGACkU2GPXarJcHWrgytyVghxbIPEvtrOP8XQf/FIBud6Z7/WFONFPSYVFkkmxCM/hOPJBj0CvG6WXV0yNN9a10lcy0yVel0JfX9g0FM0FH4H8pSKEqV2byTcoQjlKQehsuw49TzKEU5pEcdwIz5sMN2XOy8V0bHuoIyoZ54NpkVtqPAMr5MQjvluuiZnU5/6shVfJjChHfYHZQ/rRQbnJhIaKTXgfCUKjm/RrzHwMhy71upSDmhKDB2A5Z1o/pOsHqwUPDW17GBNmFDE/SnbpHhGxemnWWebfxTredFQ6YAy+zThDCXTzglSLsgi64ThDJsHN32/PEa0IiXM6moeRPZOK2NapFF8jFq8WYXvlk0Ianfl9TvgrfEufVx2o+V/0DUxo7TxeoukRuHWsJ7SGfFnWUhoj75mJxgvVLA82SdgTllPWYXTIJBUZ+XsoauOsH+VkDoNINEU3pQOySZj5dzzTYwglnLTBOP8KGxA5zLXSRleSFqwIDAQABMA0GCSqGSIb3DQEBCwUAA4ICAQAXnZZoJgo8I8zOHoQQa0Ocik9k9MCeO7M0gPtD+Xe9JRfoMolxaEZnezmADuJTCepUOUi1cgZXCScmDer2Zc2Y9pldJKhAitBiaajUrTfd0Dl8Gd8WGip8NN+8L9CsELZ+/hQnTG2GHGwi21s/yWt4yT3h2cIViuBqRvNTaxkMh1Devtzlx7haVjNCcDO5muIVBTBynJiaQV5zRaTYiLh+hT6O4OccOHJfnRdFkBRCnnCXE4qtrJg9XJ+NqkP3y0MBZueeQsdnmz9LgSwTiQHWgBI7nJSk0sLgw4AaT18xaZsx9xalKDcy7PN9Ya8IldcG4z+DP2cAoZsKejbZBfsvkV/gYC0g/LxBw0sGJrDaFc8BGDeJRxqwrpsJC8YnXFDi5/SwKII0CpOtb2MxwZC2uzmA9srnV05ta8MbdIQ2xsA8T0MDkTjOPqpJDUg9cqXZbOEOiUywJpYG6XxJdkbxx+IYyOyv8Rn0kLwwAgml7JF3w75fCDKwMw2gEY/0inqtS2NleA8XmA+CZ16YTTBQobyLUrsauVmJm4adRKFgq1OxCRbCGPeRRtD772cpAue2ZTD6Oa8UQlCkEdmXQYjp2PoguWmFI/X9T9P4oREZ182hP4b3EFr2WAhH2waURJmXVATR8IsvKxkSbBxCdVn1zhD55zuOBNLX5f4kq/4ipQ==",  # noqa: E501
        "description":
            "Example certificate for end2end tests.",
        "name": cert_name,
        "type": "secure-boot"
    }

    return console.certificates.import_certificate(cpc, props), props


class StatusError(Exception):
    """
    Indicates that a desired status cannot be reached.
    """
    pass


def ensure_lpar_inactive(lpar):
    """
    Ensure that the LPAR is inactive, regardless of what its current status is.

    If this function returns, the status of the LPAR will be 'not-activated'.

    If that status cannot be reached within the operation and status timeout
    of the session, StatusError is raised.

    Parameters:

      lpar (zhmcclient.Lpar): The LPAR (must exist).

    Raises:

      StatusError: The inactive status cannot be reached.
      zhmcclient.CeasedExistence: The LPAR no longer exists.
      zhmcclient.Error: Any zhmcclient exception can happen, except
        OperationTimeout and StatusTimeout (which result in StatusError).
    """
    org_status = pull_lpar_status(lpar)
    if org_status == 'not-activated':
        return

    try:
        lpar.deactivate(force=True)
    except zhmcclient.OperationTimeout:
        status = pull_lpar_status(lpar)
        timeout = lpar.manager.session.retry_timeout_config.operation_timeout
        raise StatusError(
            f"Could not get LPAR {lpar.name!r} from status {org_status!r} into "
            f"status 'not-activated' within operation timeout {timeout}; "
            f"current status is: {status!r}")
    except zhmcclient.StatusTimeout:
        status = pull_lpar_status(lpar)
        timeout = lpar.manager.session.retry_timeout_config.status_timeout
        raise StatusError(
            f"Could not get LPAR {lpar.name!r} from status {org_status!r} into "
            f"status 'not-activated' within status timeout {timeout}; "
            f"current status is: {status!r}")

    # This is just an additional check which should not fail.
    status = pull_lpar_status(lpar)
    if status != 'not-activated':
        raise StatusError(
            f"Could not get LPAR {lpar.name!r} from status {org_status!r} into "
            "status 'not-activated' for unknown reasons; "
            f"current status is: {status!r}")


def set_resource_property(resource, name, value):
    """
    Set a property on a zhmcclient resource to a value and return the current
    value (retrieved freshly from the HMC).

    Parameters:

      resource (zhmcclient.BaseResource): The resource (e.g. LPAR).
      name (string): Name of the property.
      value (object): New value for the property.

    Returns:

      object: Old value of the property.

    Raises:

      zhmcclient.CeasedExistence: The resource no longer exists.
      zhmcclient.Error: Any zhmcclient exception can happen, except
        OperationTimeout and StatusTimeout.
    """
    resource.pull_full_properties()
    old_value = resource.get_property(name)
    resource.update_properties({name: value})
    return old_value


def pull_lpar_status(lpar):
    """
    Retrieve the current LPAR status on the HMC as fast as possible and return
    it.

    LPAR status values and their meaning:

    Status         Resources allocated  OS running  Notes
    ---------------------------------------------------------------------------
    not-activated  no                   no
    not-operating  yes                  no          All CPUs are stopped
    operating      yes                  yes         No degradations
    exceptions     yes                  yes         Some CPUs are stopped

    Note that the status "acceptable" is only shown on the SE GUI, but will
    never appear at the WS-API.

    Raises:

      zhmcclient.CeasedExistence: The LPAR no longer exists.
      zhmcclient.Error: Any zhmcclient exception can happen.
    """
    lpars = lpar.manager.cpc.lpars.list(filter_args={'name': lpar.name})
    if len(lpars) != 1:
        raise zhmcclient.CeasedExistence(lpar.uri)
    this_lpar = lpars[0]
    actual_status = this_lpar.get_property('status')
    return actual_status


def is_cpc_property_hmc_inventory(name):
    """
    Return whether a CPC property defined in the HMC inventory file is
    actually a CPC property (vs. a special property).
    """
    special_properties = (
        'loadable_lpars',
        'load_profiles',
    )
    return name not in special_properties


def skip_missing_api_feature(
        console, console_feature, cpc=None, cpc_feature=None):
    """
    Skip pytest testcase if an HMC Console or CPC API feature is not
    available.
    """
    client = console.manager.client

    api_version_info = client.version_info()
    if api_version_info < (4, 10):
        pytest.skip(
            f"HMC API version {'.'.join(map(str, api_version_info))} is below "
            "minimum version 4.10 required for API features")

    if console_feature:
        if not console.list_api_features(console_feature):
            pytest.skip(
                f"Console API feature {console_feature!r} is not available")

    if cpc_feature:
        if not cpc.list_api_features(cpc_feature):
            pytest.skip(
                f"CPC API feature {cpc_feature!r} is not available for "
                f"CPC {cpc.name!r}")



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _virtual_function module.
"""


# FIXME: Migrate requests_mock to zhmcclient_mock.
import requests_mock

from zhmcclient import Session, Client, VirtualFunction


class TestVirtualFunction:
    """
    All tests for VirtualFunction and VirtualFunctionManager classes.
    """

    def setup_method(self):
        """
        Setup that is called by pytest before each test method.
        """
        # pylint: disable=attribute-defined-outside-init

        self.session = Session('test-dpm-host', 'test-user', 'test-id')
        self.client = Client(self.session)
        with requests_mock.mock() as m:
            # Because logon is deferred until needed, we perform it
            # explicitly in order to keep mocking in the actual test simple.
            m.post(
                '/api/sessions', json={
                    'api-session': 'test-session-id',
                    'notification-topic': 'test-obj-topic.1',
                    'job-notification-topic': 'test-job-topic.1',
                    'session-credential':
                        'un8bu462g37aw9j0o8pltontz3szt35jh4b1qe2toxt6fkhl4',
                })
            m.get(
                '/api/version', json={
                    'api-major-version': 4,
                    'api-minor-version': 10,
                })
            self.session.logon()

        self.cpc_mgr = self.client.cpcs
        with requests_mock.mock() as m:
            result = {
                'cpcs': [
                    {
                        'object-uri': '/api/cpcs/fake-cpc-id-1',
                        'name': 'CPC1',
                        'status': '',
                    }
                ]
            }
            m.get('/api/cpcs', json=result)
#            self.cpc = self.cpc_mgr.find(name="CPC1", full_properties=False)
            cpcs = self.cpc_mgr.list()
            self.cpc = cpcs[0]

        partition_mgr = self.cpc.partitions
        with requests_mock.mock() as m:
            result = {
                'partitions': [
                    {
                        'status': 'active',
                        'object-uri': '/api/partitions/fake-part-id-1',
                        'name': 'PART1'
                    },
                    {
                        'status': 'stopped',
                        'object-uri': '/api/partitions/fake-part-id-2',
                        'name': 'PART2'
                    }
                ]
            }

            m.get('/api/cpcs/fake-cpc-id-1/partitions', json=result)

            mock_result_get_bulk = [
                {
                    'id': '1',
                    'status': 200,
                    'body': {
                        'status': 'active',
                        'object-uri': '/api/partitions/fake-part-id-1',
                        'name': 'PART1',
                        'description': 'Test Partition',
                        'more_properties': 'bliblablub',
                        'virtual-function-uris': [
                            '/api/partitions/fake-part-id-1/virtual-functions/'
                            'fake-vf-id-1',
                            '/api/partitions/fake-part-id-1/virtual-functions/'
                            'fake-vf-id-2'
                        ]
                    },
                },
                {
                    'id': '2',
                    'status': 200,
                    'body': {
                        'status': 'stopped',
                        'object-uri': '/api/partitions/fake-part-id-2',
                        'name': 'PART2',
                        'description': 'Test Partition',
                        'more_properties': 'bliblablub',
                        'virtual-function-uris': [
                            '/api/partitions/fake-part-id-2/virtual-functions/'
                            'fake-vf-id-1',
                            '/api/partitions/fake-part-id-2/virtual-functions/'
                            'fake-vf-id-2'
                        ]
                    }
                }
            ]
            m.post('/api/services/aggregation/submit',
                   json=mock_result_get_bulk)

            partitions = partition_mgr.list(full_properties=True)
            self.partition = partitions[0]

    def teardown_method(self):
        """
        Teardown that is called by pytest after each test method.
        """
        with requests_mock.mock() as m:
            m.delete('/api/sessions/this-session', status_code=204)
            self.session.logoff()

    def test_init(self):
        """Test __init__() on VirtualFunctionManager instance in Partition."""
        vf_mgr = self.partition.virtual_functions
        assert vf_mgr.partition == self.partition

    def test_list_short_ok(self):
        """
        Test successful list() with short set of properties on
        VirtualFunctionManager instance in partition.
        """
        vf_mgr = self.partition.virtual_functions
        vfs = vf_mgr.list(full_properties=False)

        assert len(vfs) == \
            len(self.partition.properties['virtual-function-uris'])
        for idx, vf in enumerate(vfs):
            assert vf.properties['element-uri'] == \
                self.partition.properties['virtual-function-uris'][idx]
            assert vf.uri == \
                self.partition.properties['virtual-function-uris'][idx]
            assert not vf.full_properties
            assert vf.manager == vf_mgr

    def test_list_full_ok(self):
        """
        Test successful list() with full set of properties on
        VirtualFunctionManager instance in partition.
        """
        vf_mgr = self.partition.virtual_functions

        with requests_mock.mock() as m:

            mock_result_vf1 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf1',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-1',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-1',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-1',
                  json=mock_result_vf1)
            mock_result_vf2 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf2',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-2',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-2',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-2',
                  json=mock_result_vf2)

            vfs = vf_mgr.list(full_properties=True)

            assert len(vfs) == \
                len(self.partition.properties['virtual-function-uris'])
            for idx, vf in enumerate(vfs):
                assert vf.properties['element-uri'] == \
                    self.partition.properties['virtual-function-uris'][idx]
                assert vf.uri == \
                    self.partition.properties['virtual-function-uris'][idx]
                assert vf.full_properties
                assert vf.manager == vf_mgr

    def test_list_filter_name_ok(self):
        """
        Test successful list() with filter arguments using the 'name' property
        on a VirtualFunctionManager instance in a partition.
        """
        vf_mgr = self.partition.virtual_functions

        with requests_mock.mock() as m:

            mock_result_vf1 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf1',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-1',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-1',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-1',
                  json=mock_result_vf1)
            mock_result_vf2 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf2',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-2',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-2',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-2',
                  json=mock_result_vf2)

            filter_args = {'name': 'vf2'}
            vfs = vf_mgr.list(filter_args=filter_args)

            assert len(vfs) == 1
            vf = vfs[0]
            assert vf.name == 'vf2'
            assert vf.uri == \
                '/api/partitions/fake-part-id-1/virtual-functions/' \
                'fake-vf-id-2'
            assert vf.properties['name'] == 'vf2'
            assert vf.properties['element-id'] == 'fake-vf-id-2'
            assert vf.manager == vf_mgr

    def test_list_filter_elementid_ok(self):
        """
        Test successful list() with filter arguments using the 'element-id'
        property on a VirtualFunctionManager instance in a partition.
        """
        vf_mgr = self.partition.virtual_functions

        with requests_mock.mock() as m:

            mock_result_vf1 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf1',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-1',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-1',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-1',
                  json=mock_result_vf1)
            mock_result_vf2 = {
                'parent': '/api/partitions/fake-part-id-1',
                'name': 'vf2',
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-2',
                'class': 'virtual-function',
                'element-id': 'fake-vf-id-2',
                'description': '',
                'more_properties': 'bliblablub'
            }
            m.get('/api/partitions/fake-part-id-1/virtual-functions/'
                  'fake-vf-id-2',
                  json=mock_result_vf2)

            filter_args = {'element-id': 'fake-vf-id-2'}
            vfs = vf_mgr.list(filter_args=filter_args)

            assert len(vfs) == 1
            vf = vfs[0]
            assert vf.name == 'vf2'
            assert vf.uri == \
                '/api/partitions/fake-part-id-1/virtual-functions/' \
                'fake-vf-id-2'
            assert vf.properties['name'] == 'vf2'
            assert vf.properties['element-id'] == 'fake-vf-id-2'
            assert vf.manager == vf_mgr

    def test_create(self):
        """
        This tests the 'Create Virtual Function' operation.
        """
        vf_mgr = self.partition.virtual_functions
        with requests_mock.mock() as m:
            result = {
                'element-uri':
                    '/api/partitions/fake-part-id-1/virtual-functions/'
                    'fake-vf-id-1'
            }
            m.post('/api/partitions/fake-part-id-1/virtual-functions',
                   json=result)

            vf = vf_mgr.create(properties={})

            assert isinstance(vf, VirtualFunction)
            assert vf.properties == result
            assert vf.uri == result['element-uri']

    def test_delete(self):
        """
        This tests the 'Delete Virtual Function' operation.
        """
        vf_mgr = self.partition.virtual_functions
        vfs = vf_mgr.list(full_properties=False)
        vf = vfs[0]
        with requests_mock.mock() as m:
            m.delete('/api/partitions/fake-part-id-1/virtual-functions/'
                     'fake-vf-id-1', status_code=204)
            vf.delete()

    def test_update_properties(self):
        """
        This tests the 'Update Virtual Function Properties' operation.
        """
        vf_mgr = self.partition.virtual_functions
        vfs = vf_mgr.list(full_properties=False)
        vf = vfs[0]
        with requests_mock.mock() as m:
            m.post('/api/partitions/fake-part-id-1/virtual-functions/'
                   'fake-vf-id-1', status_code=204)
            vf.update_properties(properties={})



# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
zhmcclient.testutils - Utilities for testing against real or mocked HMCs.
"""


from ._hmc_inventory_file import *         # noqa: F401
from ._hmc_vault_file import *             # noqa: F401
from ._hmc_definition import *             # noqa: F401
from ._hmc_definitions import *            # noqa: F401
from ._hmc_definition_fixtures import *    # noqa: F401
from ._cpc_fixtures import *               # noqa: F401



# Copyright 2017,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for tests.
"""

import sys
import os
import logging
from dateutil import tz

import zhmcclient
from zhmcclient_mock._hmc import FakedBaseResource, \
    FakedMetricGroupDefinition, FakedMetricObjectValues

# Logger names by log component
LOGGER_NAMES = {
    'all': '',  # root logger
    'api': zhmcclient.API_LOGGER_NAME,
    'hmc': zhmcclient.HMC_LOGGER_NAME,
}

DEFAULT_LOG = 'all=warning'

DEFAULT_RT_CONFIG = zhmcclient.RetryTimeoutConfig(
    connect_timeout=10,
    connect_retries=1,
    operation_timeout=300,
    status_timeout=60,
)


def assert_resources(resources, exp_resources, prop_names):
    """
    Assert that a list of resource objects is equal to an expected list of
    resource objects (or faked resource objects).

    This is done by comparing:
    - The resource URIs, making sure that the two lists have matching URIs.
    - The specified list of property names.

    Parameters:

      resources (list): List of BaseResource objects to be checked.

      exp_resources (list): List of BaseResource or FakedResource objects
        defining the expected number of objects and property values.

      prop_names (list): List of property names to be checked. If `None`, all
        properties in the expected resource object are checked.
    """

    # Assert the resource URIs
    uris = {res.uri for res in resources}
    exp_uris = {res.uri for res in exp_resources}
    assert uris == exp_uris, \
        f"Unexpected URIs: got: {uris}, expected: {exp_uris}"

    for res in resources:

        # Search for the corresponding expected resource
        exp_res = None
        for _exp_res in exp_resources:
            if _exp_res.uri == res.uri:
                exp_res = _exp_res
                break
        assert exp_res is not None

        # Assert the property names to be checked
        if prop_names is None:
            _prop_names = exp_res.properties.keys()
        else:
            _prop_names = prop_names
        for prop_name in _prop_names:

            # Not all resources in a list result have all properties
            # (e.g. Crypto adapter does not have 'port-count'), so we check
            # the property on the actual resource only when the expected
            # resource has the property.
            if prop_name in exp_res.properties:

                assert prop_name in res.properties, (
                    f"Expected property {prop_name!r} missing in resource "
                    f"{res.name!r}: Resource: {res!r}")

                prop_value = res.properties[prop_name]
                exp_prop_value = exp_res.properties[prop_name]
                assert prop_value == exp_prop_value, (
                    f"Unexpected value for property {prop_name!r} in resource "
                    f"{res.name!r}: got: {prop_value!r}, expected: "
                    f"{exp_prop_value!r}, resource: {res!r}")


def info(capsys, format_str, format_args=None):
    """
    Print an information message during test execution that is not being
    captured by py.test (i.e. it is shown even when py.test is not invoked
    with '-s').

    Parameters:

      capsys: The pytest 'capsys' fixture the testcase function must specify as
        an argument and pass to this function here.

      format_str (string): Percent-based format string.

      format_args: Argument(s) for the format string.
    """

    if format_args is not None:
        msg = (format_str % format_args)
    else:
        msg = format_str

    with capsys.disabled():
        print(msg)


def print_logging():
    """
    Debug function that prints the relevant settings of all Python loggers
    that are relevant for zhmcclient.
    """
    logger_names = LOGGER_NAMES.values()
    for logger_name in logger_names:
        logger = logging.getLogger(logger_name)
        print_logger(logger)


def print_logger(logger):
    """
    Debug function that prints the relevant settings of a Python logger.
    """
    print(f"Debug: Logger {logger.name!r}:")
    print(f"Debug:   logger level: {logger.level} "
          f"({logging.getLevelName(logger.level)})")
    if not logger.handlers:
        print("Debug:   No handlers")
    for handler in logger.handlers:
        print(f"Debug:   Handler {type(handler)}:")
        print(f"Debug:     handler level: {handler.level} "
              f"({logging.getLevelName(handler.level)})")
        _fmt = getattr(handler.formatter, '_fmt', None)
        print(f"Debug:     handler format: {_fmt!r}")


def setup_logging():
    """
    Set up logging for the zhmcclient, based on the value of the ZHMC_LOG
    env. variable with the following value::

        COMP=LEVEL[,COMP=LEVEL[,...]]

    Where:

      * ``COMP`` is one of: ``all``, ``api``, ``hmc``.
      * ``LEVEL`` is one of: ``error``, ``warning``, ``info``, ``debug``.

    If the variable is not set, this defaults to::

        all=warning
    """

    log = os.environ.get('ZHMC_LOG', None)

    if log is None:
        log = DEFAULT_LOG

    log_components = LOGGER_NAMES.keys()

    for lc in log_components:
        reset_logger(lc)

    handler = logging.StreamHandler(stream=sys.stderr)
    fs = '%(levelname)s %(name)s: %(message)s'
    handler.setFormatter(logging.Formatter(fs))

    log_specs = log.split(',')
    for log_spec in log_specs:

        # ignore extra ',' at begin, end or in between
        if log_spec == '':
            continue

        try:
            log_comp, log_level = log_spec.split('=', 1)
        except ValueError:
            raise ValueError("Missing '=' in COMP=LEVEL specification "
                             f"in ZHMC_LOG variable: {log_spec}")

        level = getattr(logging, log_level.upper(), None)
        if level is None:
            raise ValueError("Invalid level in COMP=LEVEL specification "
                             f"in ZHMC_LOG variable: {log_spec}")

        if log_comp not in log_components:
            raise ValueError("Invalid component in COMP=LEVEL specification "
                             f"in ZHMC_LOG variable: {log_spec}")

        setup_logger(log_comp, handler, level)


def reset_logger(log_comp):
    """
    Reset the logger for the specified log component (unless it is the root
    logger) to add a NullHandler if it does not have any handlers. Having a
    handler prevents a log request to be propagated to the parent logger.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    if name != '' and not logger.handlers:
        logger.addHandler(logging.NullHandler())


def setup_logger(log_comp, handler, level):
    """
    Setup the logger for the specified log component to add the specified
    handler (removing a possibly present NullHandler) and to set it to the
    specified log level. The handler is also set to the specified log level
    because the default level of a handler is 0 which causes it to process all
    levels.
    """

    name = LOGGER_NAMES[log_comp]
    logger = logging.getLogger(name)

    for h in logger.handlers:
        if isinstance(h, logging.NullHandler):
            logger.removeHandler(h)

    handler.setLevel(level)
    logger.addHandler(handler)
    logger.setLevel(level)


def assert_equal_hmc(hmc1, hmc2):
    """
    Assert that two FakedHmc objects are equal.
    """
    assert hmc1.api_version == hmc2.api_version
    assert hmc1.hmc_name == hmc2.hmc_name
    assert hmc1.hmc_version == hmc2.hmc_version
    assert hmc1.enabled == hmc2.enabled

    consoles1 = hmc1.consoles.list()
    consoles2 = hmc2.consoles.list()
    assert len(consoles1) == len(consoles2)
    for i, console1 in enumerate(consoles1):
        console2 = consoles2[i]
        assert_equal_resource(console1, console2)

        users1 = console1.users.list()
        users2 = console2.users.list()
        assert len(users1) == len(users2)
        for j, user1 in enumerate(users1):
            user2 = users2[j]
            assert_equal_resource(user1, user2)

        user_roles1 = console1.user_roles.list()
        user_roles2 = console2.user_roles.list()
        assert len(user_roles1) == len(user_roles2)
        for j, user_role1 in enumerate(user_roles1):
            user_role2 = user_roles2[j]
            assert_equal_resource(user_role1, user_role2)

        user_patterns1 = console1.user_patterns.list()
        user_patterns2 = console2.user_patterns.list()
        assert len(user_patterns1) == len(user_patterns2)
        for j, user_pattern1 in enumerate(user_patterns1):
            user_pattern2 = user_patterns2[j]
            assert_equal_resource(user_pattern1, user_pattern2)

        password_rules1 = console1.password_rules.list()
        password_rules2 = console2.password_rules.list()
        assert len(password_rules1) == len(password_rules2)
        for j, password_rule1 in enumerate(password_rules1):
            password_rule2 = password_rules2[j]
            assert_equal_resource(password_rule1, password_rule2)

        tasks1 = console1.tasks.list()
        tasks2 = console2.tasks.list()
        assert len(tasks1) == len(tasks2)
        for j, task1 in enumerate(tasks1):
            task2 = tasks2[j]
            assert_equal_resource(task1, task2)

        lsds1 = console1.ldap_server_definitions.list()
        lsds2 = console2.ldap_server_definitions.list()
        assert len(lsds1) == len(lsds2)
        for j, lsd1 in enumerate(lsds1):
            lsd2 = lsds2[j]
            assert_equal_resource(lsd1, lsd2)

        # Do not compare unmanaged CPCs, since they are not dumped.

        storage_groups1 = console1.storage_groups.list()
        storage_groups2 = console2.storage_groups.list()
        assert len(storage_groups1) == len(storage_groups2)
        for j, storage_group1 in enumerate(storage_groups1):
            storage_group2 = storage_groups2[j]
            assert_equal_resource(storage_group1, storage_group2)

            storage_volumes1 = storage_group1.storage_volumes.list()
            storage_volumes2 = storage_group2.storage_volumes.list()
            assert len(storage_volumes1) == len(storage_volumes2)
            for k, storage_volume1 in enumerate(storage_volumes1):
                storage_volume2 = storage_volumes2[k]
                assert_equal_resource(storage_volume1, storage_volume2)

    cpcs1 = hmc1.cpcs.list()
    cpcs2 = hmc1.cpcs.list()
    assert len(cpcs1) == len(cpcs2)
    for i, cpc1 in enumerate(cpcs1):
        cpc2 = cpcs2[i]
        assert_equal_resource(cpc1, cpc2)

        cgs1 = cpc1.capacity_groups.list()
        cgs2 = cpc2.capacity_groups.list()
        assert len(cgs1) == len(cgs2)
        for j, cg1 in enumerate(cgs1):
            cg2 = cgs2[j]
            assert_equal_resource(cg1, cg2)

        partitions1 = cpc1.partitions.list()
        partitions2 = cpc2.partitions.list()
        assert len(partitions1) == len(partitions2)
        for j, partition1 in enumerate(partitions1):
            partition2 = partitions2[j]
            assert_equal_resource(partition1, partition2)

            nics1 = partition1.nics.list()
            nics2 = partition2.nics.list()
            assert len(nics1) == len(nics2)
            for k, nic1 in enumerate(nics1):
                nic2 = nics2[k]
                assert_equal_resource(nic1, nic2)

            if partition1.hbas is not None and partition2.hbas is not None:
                # z14 and above do not have .hbas set
                hbas1 = partition1.hbas.list()
                hbas2 = partition2.hbas.list()
                assert len(hbas1) == len(hbas2)
                for k, hba1 in enumerate(hbas1):
                    hba2 = hbas2[k]
                    assert_equal_resource(hba1, hba2)

            vfs1 = partition1.virtual_functions.list()
            vfs2 = partition2.virtual_functions.list()
            assert len(vfs1) == len(vfs2)
            for k, vf1 in enumerate(vfs1):
                vf2 = vfs2[k]
                assert_equal_resource(vf1, vf2)

        adapters1 = cpc1.adapters.list()
        adapters2 = cpc2.adapters.list()
        assert len(adapters1) == len(adapters2)
        for j, adapter1 in enumerate(adapters1):
            adapter2 = adapters2[j]
            assert_equal_resource(adapter1, adapter2)

            if adapter1.properties.get('type', None) != 'not-configured':
                # Unconfigured FICON adapters cannot retrieve port properties
                ports1 = adapter1.ports.list()
                ports2 = adapter2.ports.list()
                assert len(ports1) == len(ports2)
                for k, port1 in enumerate(ports1):
                    port2 = ports2[k]
                    assert_equal_resource(port1, port2)

        vss1 = cpc1.virtual_switches.list()
        vss2 = cpc2.virtual_switches.list()
        assert len(vss1) == len(vss2)
        for j, vs1 in enumerate(vss1):
            vs2 = vss2[j]
            assert_equal_resource(vs1, vs2)

        lpars1 = cpc1.lpars.list()
        lpars2 = cpc2.lpars.list()
        assert len(lpars1) == len(lpars2)
        for j, lpar1 in enumerate(lpars1):
            lpar2 = lpars2[j]
            assert_equal_resource(lpar1, lpar2)

        raps1 = cpc1.reset_activation_profiles.list()
        raps2 = cpc2.reset_activation_profiles.list()
        assert len(raps1) == len(raps2)
        for j, rap1 in enumerate(raps1):
            rap2 = raps2[j]
            assert_equal_resource(rap1, rap2)

        iaps1 = cpc1.image_activation_profiles.list()
        iaps2 = cpc2.image_activation_profiles.list()
        assert len(iaps1) == len(iaps2)
        for j, iap1 in enumerate(iaps1):
            iap2 = iaps2[j]
            assert_equal_resource(iap1, iap2)

        laps1 = cpc1.load_activation_profiles.list()
        laps2 = cpc2.load_activation_profiles.list()
        assert len(laps1) == len(laps2)
        for j, lap1 in enumerate(laps1):
            lap2 = laps2[j]
            assert_equal_resource(lap1, lap2)

    # TODO: Reactivate metrics comparison once implemented
    # mcs1 = hmc1.metrics_contexts.list()
    # mcs2 = hmc2.metrics_contexts.list()
    # assert len(mcs1) == len(mcs2)
    # for i, mc1 in enumerate(mcs1):
    #     mc2 = mcs2[i]
    #     assert_equal_resource(mc1, mc2)
    # mgd1_mg_names = hmc1.metrics_contexts.get_metric_group_definition_names()
    # mgd2_mg_names = hmc2.metrics_contexts.get_metric_group_definition_names()
    # assert set(mgd1_mg_names) == set(mgd2_mg_names)
    # for mg_name in mgd1_mg_names:
    #     mgd1 = hmc1.metrics_contexts.get_metric_group_definition(mg_name)
    #     mgd2 = hmc2.metrics_contexts.get_metric_group_definition(mg_name)
    #     assert_equal_metric_group_def(mgd1, mgd2)
    # mv1_mg_names = hmc1.metrics_contexts.get_metric_values_group_names()
    # mv1_mg_names = hmc2.metrics_contexts.get_metric_values_group_names()
    # assert set(mv1_mg_names) == set(mv1_mg_names)
    # for mg_name in mv1_mg_names:
    #     mv1_list = hmc1.metrics_contexts.get_metric_values(mg_name)
    #     mv2_list = hmc2.metrics_contexts.get_metric_values(mg_name)
    #     for i, mv1 in enumerate(mv1_list):
    #         mv2 = mv2_list[i]
    #         assert_equal_metric_values(mv1, mv2)


def assert_equal_resource(res1, res2):
    """
    Assert that two Faked resource objects are equal.

    Only the standard attributes are compared.
    """
    assert isinstance(res1, FakedBaseResource)
    assert isinstance(res2, FakedBaseResource)
    assert res1.uri == res2.uri
    assert res1.oid == res2.oid
    names1 = set(res1.properties.keys())
    names2 = set(res2.properties.keys())
    if names1 != names2:
        raise AssertionError(
            "Resources do not have the same set of properties:\n"
            f"- res1 names: {sorted(names1)}\n"
            f"- res2 names: {sorted(names2)}\n")
    for name in res1.properties:
        value1 = res1.properties[name]
        value2 = res2.properties[name]
        if value1 != value2:
            raise AssertionError(
                f"Resources do not have the same value for property {name}:\n"
                f"- res1 value: {value1}\n"
                f"- res2 value: {value2}\n")


def assert_equal_metric_group_def(mgd1, mgd2):
    """
    Assert that two FakedMetricGroupDefinition objects are equal.
    """
    assert isinstance(mgd1, FakedMetricGroupDefinition)
    assert isinstance(mgd2, FakedMetricGroupDefinition)
    assert mgd1.name == mgd2.name
    assert mgd1.types == mgd2.types


def assert_equal_metric_values(mv1, mv2):
    """
    Assert that two FakedMetricObjectValues objects are equal.
    """
    assert isinstance(mv1, FakedMetricObjectValues)
    assert isinstance(mv2, FakedMetricObjectValues)
    assert mv1.group_name == mv2.group_name
    assert mv1.resource_uri == mv2.resource_uri
    assert mv1.timestamp == mv2.timestamp
    assert mv1.values == mv2.values


def timestamp_aware(dt):
    """
    Return the input datetime object as a timezone-aware datetime object.

    If the input object is already timezone-aware, it is returned.

    If the input object is timezone-naive, the local timezone information is
    added to a copy of the input object and that copy is returned.
    """
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=tz.tzlocal())  # new object
    return dt


def assert_blanked_in_message(message, properties, blanked_properties):
    """
    Assert that a message containing a string representation of a properties
    dict has the desired properties blanked out.

    Parameters:
        message (str): The message to be checked.
        properties (dict): Properties that can possibly be in the message
          (with hyphened names).
        blanked_properties (list of str): The names of the properties that
          need to be blanked out in the message (with hyphened names).
    """
    for pname in blanked_properties:
        if pname in properties:
            exp_str = f"'{pname}': '{zhmcclient.BLANKED_OUT_STRING}'"
            assert message.find(exp_str) > 0



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An :term:`Activation Profile` controls the activation of a :term:`CPC`
or :term:`LPAR`. They are used to tailor the operation of a CPC and are
stored in the Support Element associated with the CPC.

Activation Profile resources are contained in CPC resources.

Activation Profile resources only exist in CPCs that are not in DPM mode.

TODO: If Reset Activation Profiles are used to determine the CPC mode,
      should they not exist in all CPC modes?

There are three types of Activation Profiles:

1. Reset:
   The Reset Activation Profile defines for a CPC the mode in which the CPC
   licensed internal code will be loaded (e.g. DPM mode or classic mode) and
   how much central storage and expanded storage will be used.

2. Image:
   For CPCs in classic mode, each LPAR can have an Image Activation Profile.
   The Image Activation Profile determines the number of CPs that the LPAR will
   use and whether these CPs will be dedicated to the LPAR or shared. It also
   allows assigning the amount of central storage and expanded storage that
   will be used by each LPAR.

3. Load:
   For CPCs in classic mode, each LPAR can have a Load Activation Profile.
   The Load Activation Profile defines the channel address of the device that
   the operating system for that LPAR will be loaded (booted) from.
"""


import copy
import warnings

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_RESET_ACTIVATION_PROFILE, RC_IMAGE_ACTIVATION_PROFILE, \
    RC_LOAD_ACTIVATION_PROFILE

__all__ = ['ActivationProfileManager', 'ActivationProfile']

# Resource class names, by profile type:
ACTIVATION_PROFILE_CLASSES = {
    'reset': RC_RESET_ACTIVATION_PROFILE,
    'image': RC_IMAGE_ACTIVATION_PROFILE,
    'load': RC_LOAD_ACTIVATION_PROFILE,
}


class ActivationProfileManager(BaseManager):
    """
    Manager providing access to the
    :term:`Activation Profiles <Activation Profile>` of a particular type in
    a particular :term:`CPC` (the scoping CPC).

    Possible types of activation profiles are:

    * Reset Activation Profile
    * Image Activation Profile
    * Load Activation Profile

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variables of a
    :class:`~zhmcclient.Cpc` object (in classic mode or ensemble mode):

    * :attr:`~zhmcclient.Cpc.reset_activation_profiles`
    * :attr:`~zhmcclient.Cpc.image_activation_profiles`
    * :attr:`~zhmcclient.Cpc.load_activation_profiles`

    HMC/SE version requirements: None
    """

    def __init__(self, cpc, profile_type):
        # This function should not go into the docs.
        # Parameters:
        #   cpc (:class:`~zhmcclient.Cpc`):
        #     CPC defining the scope for this manager.
        #   profile_type (string):
        #     Type of Activation Profiles:
        #     * `reset`: Reset Activation Profiles
        #     * `image`: Image Activation Profiles
        #     * `load`: Load Activation Profiles

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name',
        ]

        try:
            activation_profile_class = ACTIVATION_PROFILE_CLASSES[profile_type]
        except KeyError:
            raise ValueError(f"Unknown activation profile type: {profile_type}")

        super().__init__(
            resource_class=ActivationProfile,
            class_name=activation_profile_class,
            session=cpc.manager.session,
            parent=cpc,
            base_uri=f'{cpc.uri}/{profile_type}-activation-profiles',
            oid_prop='name',  # This is an exception!
            uri_prop='element-uri',
            name_prop='name',
            query_props=query_props,
            supports_properties=True)

        self._profile_type = profile_type

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: :term:`CPC` defining the scope for this
        manager.
        """
        return self._parent

    @property
    def profile_type(self):
        """
        :term:`string`: Type of the Activation Profiles managed by this object:

        * ``'reset'`` - Reset Activation Profiles
        * ``'image'`` - Image Activation Profiles
        * ``'load'`` - Load Activation Profiles
        """
        return self._profile_type

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the Activation Profiles of this CPC, of the profile type
        managed by this object.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the HMC List operation is performed with the subset of the
          provided filter arguments that can be handled on the HMC side and the
          remaining filter arguments are applied on the client side on the list
          result.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to this CPC.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

            This parameter requires HMC 2.16.0 or higher, and is supported
            only for image profiles.

        Returns:

          : A list of :class:`~zhmcclient.ActivationProfile` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = self._profile_type + '-activation-profiles'
        list_uri = f'{self.cpc.uri}/{result_prop}'
        if self._profile_type != 'image' and additional_properties is not None:
            raise TypeError(
                f"list() for {self._profile_type} profiles does not support "
                "'additional_properties' parameter")
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call(blanked_properties=['ssc-master-pw', 'zaware-master-pw'],
                     properties_pos=1)
    def create(self, properties):
        """
        Create and configure an Activation Profiles on this CPC, of the profile
        type managed by this object.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "create-delete-activation-profiles"

        Authorization requirements:

        * Object-access permission to this CPC.
        * Task permission to the "Customize/Delete Activation Profiles" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create Reset/Image/Load Activation Profile' in the
            :term:`HMC API` book.

            Note that the input profile name for creation must be provided in
            property 'profile-name', even though it shows up on the created
            resource in property 'name'. This applies to all three types of
            activation profiles.

        Returns:

          ActivationProfile:
            The resource object for the new Activation Profile.
            The object will have its 'element-uri' property set, and will also
            have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        ap_selector = self._profile_type + '-activation-profiles'
        uri = f'{self.cpc.uri}/{ap_selector}'

        result = self.session.post(uri, body=properties)

        # The "Create ... Activation Profile" operations do not return the
        # resource URI, so we construct it ourselves. Also, these operations
        # specify the profile name in input property 'profile-name'.
        if result is not None:
            warnings.warn(
                f"The Create {self._profile_type} Activation Profile operation "
                f"now has response data with properties: {result.keys()!r}",
                UserWarning)
        name = properties['profile-name']
        uri = f'{uri}/{name}'

        props = copy.deepcopy(properties)
        props[self._uri_prop] = uri
        profile = ActivationProfile(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return profile


class ActivationProfile(BaseResource):
    """
    Representation of an :term:`Activation Profile` of a particular type.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.ActivationProfileManager`).

    HMC/SE version requirements: None
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.ActivationProfileManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, ActivationProfileManager), (
            "ActivationProfile init: Expected manager type "
            f"{ActivationProfileManager}, got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this Activation Profile.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "create-delete-activation-profiles"

        Authorization requirements:

        * Task permission to the "Customize/Delete Activation Profiles" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call(blanked_properties=['ssc-master-pw', 'zaware-master-pw'],
                     properties_pos=1)
    def update_properties(self, properties):
        """
        Update writeable properties of this Activation Profile.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements: None

        Authorization requirements:

        * Object-access permission to the CPC of this Activation Profile.
        * Task permission for the "Customize/Delete Activation Profiles" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section
            '<profile_type> activation profile' in the :term:`HMC API` book,
            where <profile_type> is the profile type of this object
            (e.g. Reset, Load, Image).

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        # Attempts to change the 'name' property will be rejected by the HMC,
        # so we don't need to update the name-to-URI cache.
        assert self.manager._name_prop not in properties
        self.update_properties_local(copy.deepcopy(properties))

    @logged_api_call
    def assign_certificate(self, certificate):
        """
        Assigns a :term:`Certificate` to this Image Activation Profile.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates".

        Authorization requirements:

        * Object-access permission to this Activation Profile.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be assigned. The certificate must not currently
            be assigned to this LPAR.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/assign-certificate', resource=self,
            body=body)

    @logged_api_call
    def unassign_certificate(self, certificate):
        """
        Unassign a :term:`Certificate` from this Image Activation Profile.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates".

        Authorization requirements:

        * Object-access permission to this Image Activation Profile.
        * Object-access permission to the specified certificate.
        * Task permission to the "Assign Secure Boot Certificates" task.

        Parameters:

          certificate (:class:`~zhmcclient.Certificate`):
            Certificate to be unassigned. The certificate must currently be
            assigned to this LPAR.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'certificate-uri': certificate.uri}
        self.manager.session.post(
            self.uri + '/operations/unassign-certificate', resource=self,
            body=body)
